# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os

import mock
from OpenSSL import crypto
from oslo_serialization import jsonutils

from vmware_nsxlib.tests.unit.v3 import mocks
from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib.tests.unit.v3 import test_client
from vmware_nsxlib.tests.unit.v3 import test_constants as const
from vmware_nsxlib.v3 import client
from vmware_nsxlib.v3 import client_cert
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import trust_management as tm


class DummyStorageDriver(dict):
    """Storage driver simulation - just a dictionary"""

    def store_cert(self, project_id, certificate, private_key):
        self[project_id] = {}
        self[project_id]['cert'] = certificate
        self[project_id]['key'] = private_key

    def get_cert(self, project_id):
        if project_id not in self:
            return (None, None)

        return (self[project_id]['cert'], self[project_id]['key'])

    def delete_cert(self, project_id):
        del(self[project_id])

    def is_empty(self, project_id):
        return project_id not in self


class NsxV3ClientCertificateTestCase(nsxlib_testcase.NsxClientTestCase):

    identity = 'drumknott'
    cert_id = "00000000-1111-2222-3333-444444444444"
    identity_id = "55555555-6666-7777-8888-999999999999"
    node_id = "meh"

    def _get_mocked_response(self, status_code, results):
        return mocks.MockRequestsResponse(
            status_code,
            jsonutils.dumps({'results': results}))

    def _get_mocked_error_response(self, status_code, error_code):
        return mocks.MockRequestsResponse(
            status_code,
            jsonutils.dumps({'httpStatus': 'go away',
                             'error_code': error_code,
                             'module_name': 'never mind',
                             'error_message': 'bad luck'}))

    def _get_mocked_trust(self, action, cert_pem):

        fake_responses = []
        if 'create' in action:
            # import cert and return its id
            results = [{'id': self.cert_id}]
            fake_responses.append(self._get_mocked_response(201, results))
            # and then bind this id to principal identity
            fake_responses.append(self._get_mocked_response(201, []))

        if 'delete' in action:
            # get certs list, including same cert imported twice edge case
            results = [{'resource_type': 'Certificate',
                        'id': 'dont care',
                        'pem_encoded': 'some junk'},
                       {'resource_type': 'Certificate',
                        'id': 'some_other_cert_id',
                        'pem_encoded': cert_pem},
                       {'resource_type': 'Certificate',
                        'id': self.cert_id,
                        'pem_encoded': cert_pem}]
            fake_responses.append(self._get_mocked_response(200, results))

            # get principal identities list
            results = [{'resource_type': 'Principal Identity',
                        'id': 'dont care',
                        'name': 'willikins',
                        'certificate_id': 'some other id'},
                       {'resource_type': 'Principal Identity',
                        'id': self.identity_id,
                        'name': self.identity,
                        'certificate_id': self.cert_id}]
            fake_responses.append(self._get_mocked_response(200, results))
            # delete certificate
            fake_responses.append(self._get_mocked_response(204, []))
            # delete identity
            fake_responses.append(self._get_mocked_response(204, []))

        mock_client = self.new_mocked_client(
            client.JSONRESTClient,
            url_prefix='api/v1', session_response=fake_responses)

        return tm.NsxLibTrustManagement(mock_client, {})

    def _verify_backend_create(self, mocked_trust, cert_pem):
        """Verify API calls to create cert and identity on backend"""
        # verify API call to import cert on backend
        base_uri = 'https://1.2.3.4/api/v1/trust-management'
        uri = base_uri + '/certificates?action=import'
        expected_body = {'pem_encoded': cert_pem}
        test_client.assert_json_call('post', mocked_trust.client, uri,
                                     single_call=False,
                                     data=jsonutils.dumps(expected_body))

        # verify API call to bind cert to identity on backend
        uri = base_uri + '/principal-identities'
        expected_body = {'name': self.identity,
                         'node_id': self.node_id,
                         'permission_group': 'read_write_api_users',
                         'certificate_id': self.cert_id,
                         'is_protected': True}
        test_client.assert_json_call('post', mocked_trust.client, uri,
                                     single_call=False,
                                     data=jsonutils.dumps(expected_body,
                                                          sort_keys=True))

    def _verify_backend_delete(self, mocked_trust):
        """Verify API calls to fetch and delete cert and identity"""
        # verify API call to query identities in order to get cert id
        base_uri = 'https://1.2.3.4/api/v1/trust-management'
        uri = base_uri + '/principal-identities'
        test_client.assert_json_call('get', mocked_trust.client, uri,
                                     single_call=False)
        # verify API call to delete openstack principal identity
        uri = uri + '/' + self.identity_id
        test_client.assert_json_call('delete', mocked_trust.client, uri,
                                     single_call=False)

        # verify API call to delete certificate
        uri = base_uri + '/certificates/' + self.cert_id
        test_client.assert_json_call('delete', mocked_trust.client, uri,
                                     single_call=False)

    def test_generate_cert(self):
        """Test startup without certificate + certificate generation"""

        storage_driver = DummyStorageDriver()
        # Prepare fake trust management for "cert create" requests
        cert_pem, key_pem = storage_driver.get_cert(self.identity)
        mocked_trust = self._get_mocked_trust('create', cert_pem)
        cert = client_cert.ClientCertificateManager(self.identity,
                                                    mocked_trust,
                                                    storage_driver)
        self.assertFalse(cert.exists())

        cert.generate(subject={}, key_size=2048, valid_for_days=333,
                      node_id=self.node_id)

        # verify client cert was generated and makes sense
        self.assertTrue(cert.exists())
        self.assertEqual(332, cert.expires_in_days())
        cert_pem, key_pem = cert.get_pem()

        # verify cert ans PK were stored in storage
        stored_cert, stored_key = storage_driver.get_cert(self.identity)
        self.assertEqual(cert_pem, stored_cert)
        self.assertEqual(key_pem, stored_key)

        # verify backend API calls
        self._verify_backend_create(mocked_trust, cert_pem)

        # try to generate cert again and fail
        self.assertRaises(nsxlib_exc.ObjectAlreadyExists,
                          cert.generate, {})

    def _prepare_storage_with_existing_cert(self, key_size, days, alg, subj):
        # prepare storage driver with existing cert and key
        # this test simulates system startup
        cert, key = client_cert.generate_self_signed_cert_pair(key_size, days,
                                                               alg, subj)
        storage_driver = DummyStorageDriver()
        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
        key_pem = crypto.dump_privatekey(crypto.FILETYPE_PEM, key)

        storage_driver.store_cert(self.identity, cert_pem, key_pem)
        return storage_driver

    def test_load_and_delete_existing_cert(self):
        """Test startup with existing certificate + certificate deletion"""

        storage_driver = self._prepare_storage_with_existing_cert(4096,
                                                                  3650,
                                                                  'sha256',
                                                                  {})

        # get mocked backend driver for trust management,
        # prepared for get request, that preceeds delete operation
        cert_pem, key_pem = storage_driver.get_cert(self.identity)
        mocked_trust = self._get_mocked_trust('delete', cert_pem)

        cert = client_cert.ClientCertificateManager(self.identity,
                                                    mocked_trust,
                                                    storage_driver)
        self.assertTrue(cert.exists())

        cert.delete()

        self.assertFalse(cert.exists())
        self.assertTrue(storage_driver.is_empty(self.identity))

        self._verify_backend_delete(mocked_trust)

    def _test_import_and_delete_cert(self, with_pkey=True):
        filename = '/tmp/test.pem'
        # this driver simulates storage==none scenario
        noop_driver = DummyStorageDriver()
        cert, key = client_cert.generate_self_signed_cert_pair(4096,
                                                               20,
                                                               'sha256',
                                                               {})

        cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
        key_pem = crypto.dump_privatekey(crypto.FILETYPE_PEM, key)
        with open(filename, 'wb') as f:
            f.write(cert_pem)
            if with_pkey:
                f.write(key_pem)

        mocked_trust = self._get_mocked_trust('create_delete',
                                              cert_pem)
        cert = client_cert.ClientCertificateManager(self.identity,
                                                    mocked_trust,
                                                    noop_driver)
        cert.import_pem(filename, self.node_id)
        self._verify_backend_create(mocked_trust, cert_pem)

        cert.delete_pem(filename)
        self._verify_backend_delete(mocked_trust)

        os.remove(filename)

    def test_import_and_delete_cert_pkey(self):
        self._test_import_and_delete_cert(True)

    def test_import_and_delete_cert_only(self):
        self._test_import_and_delete_cert(False)

    def test_get_certificate_details(self):
        """Test retrieving cert details for existing cert"""

        key_size = 2048
        days = 999
        alg = 'sha256'
        subj = {'country': 'CA',
                'organization': 'squirrel rights',
                'hostname': 'www.squirrels.ca',
                'unit': 'nuts',
                'state': 'BC'}

        storage_driver = self._prepare_storage_with_existing_cert(key_size,
                                                                  days, alg,
                                                                  subj)
        with client_cert.ClientCertificateManager(self.identity,
                                                  None,
                                                  storage_driver) as cert:
            self.assertTrue(cert.exists())
            self.assertEqual(days - 1, cert.expires_in_days())
            self.assertEqual(key_size, cert.get_key_size())
            cert_subj = cert.get_subject()
            self.assertEqual(subj, cert_subj)

    def test_bad_certificate_values(self):
        bad_cert_values = [{'key_size': 1024,
                            'valid_for_days': 10,
                            'signature_alg': 'sha256',
                            'subject': {}},
                           {'key_size': 4096,
                            'valid_for_days': 100,
                            'signature_alg': 'sha',
                            'subject': {}}]

        for args in bad_cert_values:
            self.assertRaises(nsxlib_exc.NsxLibInvalidInput,
                              client_cert.generate_self_signed_cert_pair,
                              **args)

    def test_find_cert_with_pem(self):
        with mock.patch.object(self.nsxlib.trust_management, 'get_certs'
                               ) as mock_get_certs:
            mock_get_certs.return_value = const.FAKE_CERT_LIST
            cert_ids = self.nsxlib.trust_management.find_cert_with_pem(
                const.FAKE_CERT_PEM)
            self.assertEqual(const.FAKE_CERT_LIST[1]['id'], cert_ids[0])



# Copyright 2015 OpenStack Foundation

# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import uuid

from oslo_log import log

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants as consts


LOG = log.getLogger(__name__)


class NSGroupManager(object):
    """This class assists with NSX integration for Neutron security-groups

    Each Neutron security-group is associated with NSX NSGroup object.
    Some specific security policies are the same across all security-groups,
    i.e - Default drop rule, DHCP. In order to bind these rules to all
    NSGroups (security-groups), we create a nested NSGroup (which its members
    are also of type NSGroups) to group the other NSGroups and associate it
    with these rules.
    In practice, one NSGroup (nested) can't contain all the other NSGroups, as
    it has strict size limit. To overcome the limited space challenge, we
    create several nested groups instead of just one, and we evenly distribute
    NSGroups (security-groups) between them.
    By using an hashing function on the NSGroup uuid we determine in which
    group it should be added, and when deleting an NSGroup (security-group) we
    use the same procedure to find which nested group it was added.
    """
    NESTED_GROUP_NAME = 'OS Nested Group'
    NESTED_GROUP_DESCRIPTION = ('OpenStack NSGroup. Do not delete.')

    def __init__(self, nsxlib, size):
        self.nsxlib_nsgroup = nsxlib.ns_group
        self._nested_groups = self._init_nested_groups(size)
        self._size = len(self._nested_groups)

    @property
    def size(self):
        return self._size

    @property
    def nested_groups(self):
        return self._nested_groups

    def _init_nested_groups(self, requested_size):
        # Construct the groups dict -
        # {0: <groups-1>,.., n-1: <groups-n>}
        size = requested_size
        nested_groups = {
            self._get_nested_group_index_from_name(nsgroup): nsgroup['id']
            for nsgroup in self.nsxlib_nsgroup.list()
            if self.nsxlib_nsgroup.is_internal_resource(nsgroup)}

        if nested_groups:
            size = max(requested_size, max(nested_groups) + 1)
            if size > requested_size:
                LOG.warning("Lowering the value of "
                            "nsx_v3:number_of_nested_groups isn't "
                            "supported, '%s' nested-groups will be used.",
                            size)

        absent_groups = set(range(size)) - set(nested_groups.keys())
        if absent_groups:
            LOG.warning(
                "Found %(num_present)s Nested Groups, "
                "creating %(num_absent)s more.",
                {'num_present': len(nested_groups),
                 'num_absent': len(absent_groups)})
            for i in absent_groups:
                cont = self._create_nested_group(i)
                nested_groups[i] = cont['id']

        return nested_groups

    def _get_nested_group_index_from_name(self, nested_group):
        # The name format is "Nested Group <index+1>"
        return int(nested_group['display_name'].split()[-1]) - 1

    def _create_nested_group(self, index):
        name_prefix = NSGroupManager.NESTED_GROUP_NAME
        name = '%s %s' % (name_prefix, index + 1)
        description = NSGroupManager.NESTED_GROUP_DESCRIPTION
        tags = self.nsxlib_nsgroup.build_v3_api_version_tag()
        return self.nsxlib_nsgroup.create(name, description, tags)

    def _hash_uuid(self, internal_id):
        return hash(uuid.UUID(internal_id))

    def _suggest_nested_group(self, internal_id):
        # Suggests a nested group to use, can be iterated to find alternative
        # group in case that previous suggestions did not help.

        index = self._hash_uuid(internal_id) % self.size
        yield self.nested_groups[index]

        for i in range(1, self.size):
            index = (index + 1) % self.size
            yield self.nested_groups[index]

    def add_nsgroup(self, nsgroup_id):
        for group in self._suggest_nested_group(nsgroup_id):
            try:
                LOG.debug("Adding NSGroup %s to nested group %s",
                          nsgroup_id, group)
                self.nsxlib_nsgroup.add_members(
                    group, consts.NSGROUP, [nsgroup_id])
                break
            except exceptions.NSGroupIsFull:
                LOG.debug("Nested group %(group_id)s is full, trying the "
                          "next group..", {'group_id': group})
        else:
            raise exceptions.ManagerError(
                details=_("Reached the maximum supported amount of "
                          "security groups."))

    def remove_nsgroup(self, nsgroup_id):
        for group in self._suggest_nested_group(nsgroup_id):
            try:
                self.nsxlib_nsgroup.remove_member(
                    group, consts.NSGROUP,
                    nsgroup_id, verify=True)
                break
            except exceptions.NSGroupMemberNotFound:
                LOG.warning("NSGroup %(nsgroup)s was expected to be found "
                            "in group %(group_id)s, but wasn't. "
                            "Looking in the next group..",
                            {'nsgroup': nsgroup_id, 'group_id': group})
                continue
        else:
            LOG.warning("NSGroup %s was marked for removal, but its "
                        "reference is missing.", nsgroup_id)



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import copy

import eventlet
import mock

from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from vmware_nsxlib.tests.unit.v3 import mocks
from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib.tests.unit.v3 import test_client
from vmware_nsxlib.tests.unit.v3 import test_constants
from vmware_nsxlib.v3 import core_resources
from vmware_nsxlib.v3 import exceptions
from vmware_nsxlib.v3 import nsx_constants
from vmware_nsxlib.v3 import resources
from vmware_nsxlib.v3 import utils


class BaseTestResource(nsxlib_testcase.NsxClientTestCase):
    """Base class for resources tests

    Contains tests for the simple get/list/delete apis
    and an api to get the mocked resource
    """
    def setUp(self, resource=None):
        self.resource = resource
        super(BaseTestResource, self).setUp()

    def get_mocked_resource(self, mock_validate=True, response=None,
                            response_repeat=1):
        session_response = None
        if response:
            session_response = mocks.MockRequestsResponse(
                200, jsonutils.dumps(response))
            if response_repeat > 1:
                session_response = [session_response] * response_repeat

        return self.mocked_resource(
            self.resource, mock_validate=mock_validate,
            session_response=session_response)

    def test_get_resource(self):
        if not self.resource:
            return
        mocked_resource = self.get_mocked_resource()
        fake_uuid = uuidutils.generate_uuid()
        mocked_resource.get(fake_uuid)
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s' % (mocked_resource.uri_segment,
                                              fake_uuid),
            headers=self.default_headers())

    def test_list_all(self):
        if not self.resource:
            return
        mocked_resource = self.get_mocked_resource()
        mocked_resource.list()
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s' % mocked_resource.uri_segment,
            headers=self.default_headers())

    def test_delete_resource(self, extra_params=None):
        if not self.resource:
            return
        mocked_resource = self.get_mocked_resource()
        fake_uuid = uuidutils.generate_uuid()
        mocked_resource.delete(fake_uuid)
        uri = 'https://1.2.3.4/api/v1/%s/%s' % (mocked_resource.uri_segment,
                                                fake_uuid)
        if extra_params:
            uri = uri + '?' + extra_params
        test_client.assert_json_call(
            'delete', mocked_resource, uri,
            headers=self.default_headers())


class TestSwitchingProfileTestCase(BaseTestResource):

    def setUp(self):
        self.types = resources.SwitchingProfileTypes
        super(TestSwitchingProfileTestCase, self).setUp(
            resources.SwitchingProfile)

    def test_switching_profile_create(self):
        mocked_resource = self.get_mocked_resource()

        mocked_resource.create(self.types.PORT_MIRRORING,
                               'pm-profile', 'port mirror prof')

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles',
            data=jsonutils.dumps({
                'resource_type': self.types.PORT_MIRRORING,
                'display_name': 'pm-profile',
                'description': 'port mirror prof'
            }, sort_keys=True),
            headers=self.default_headers())

    def test_switching_profile_update(self):

        tags = [
            {
                'scope': 'os-project-id',
                'tag': 'project-1'
            },
            {
                'scope': 'os-api-version',
                'tag': '2.1.1.0'
            }
        ]

        mocked_resource = self.get_mocked_resource()
        fake_uuid = uuidutils.generate_uuid()

        mocked_resource.update(
            fake_uuid, self.types.PORT_MIRRORING, tags=tags)

        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles/%s' % fake_uuid,
            data=jsonutils.dumps({
                'resource_type': self.types.PORT_MIRRORING,
                'tags': tags
            }, sort_keys=True),
            headers=self.default_headers())

    def test_spoofgaurd_profile_create(self):

        tags = [
            {
                'scope': 'os-project-id',
                'tag': 'project-1'
            },
            {
                'scope': 'os-api-version',
                'tag': '2.1.1.0'
            }
        ]

        mocked_resource = self.get_mocked_resource()

        mocked_resource.create_spoofguard_profile(
            'plugin-spoof', 'spoofguard-for-plugin',
            whitelist_ports=True, tags=tags)

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles',
            data=jsonutils.dumps({
                'resource_type': self.types.SPOOF_GUARD,
                'display_name': 'plugin-spoof',
                'description': 'spoofguard-for-plugin',
                'white_list_providers': ['LPORT_BINDINGS'],
                'tags': tags
            }, sort_keys=True),
            headers=self.default_headers())

    def test_create_dhcp_profile(self):

        tags = [
            {
                'scope': 'os-project-id',
                'tag': 'project-1'
            },
            {
                'scope': 'os-api-version',
                'tag': '2.1.1.0'
            }
        ]

        mocked_resource = self.get_mocked_resource()

        mocked_resource.create_dhcp_profile(
            'plugin-dhcp', 'dhcp-for-plugin',
            tags=tags)

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles',
            data=jsonutils.dumps({
                'bpdu_filter': {
                    'enabled': True,
                    'white_list': []
                },
                'resource_type': self.types.SWITCH_SECURITY,
                'display_name': 'plugin-dhcp',
                'description': 'dhcp-for-plugin',
                'tags': tags,
                'dhcp_filter': {
                    'client_block_enabled': True,
                    'server_block_enabled': False
                },
                'rate_limits': {
                    'enabled': False,
                    'rx_broadcast': 0,
                    'tx_broadcast': 0,
                    'rx_multicast': 0,
                    'tx_multicast': 0
                },
                'block_non_ip_traffic': True
            }, sort_keys=True),
            headers=self.default_headers())

    def test_create_mac_learning_profile(self):

        tags = [
            {
                'scope': 'os-project-id',
                'tag': 'project-1'
            },
            {
                'scope': 'os-api-version',
                'tag': '2.1.1.0'
            }
        ]

        mocked_resource = self.get_mocked_resource()

        mocked_resource.create_mac_learning_profile(
            'plugin-mac-learning', 'mac-learning-for-plugin',
            tags=tags)

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles',
            data=jsonutils.dumps({
                'mac_learning': {
                    'enabled': True,
                },
                'resource_type': self.types.MAC_LEARNING,
                'display_name': 'plugin-mac-learning',
                'description': 'mac-learning-for-plugin',
                'tags': tags,
                'mac_change_allowed': True,
            }, sort_keys=True),
            headers=self.default_headers())

    def test_find_by_display_name(self):
        resp_resources = {
            'results': [
                {'display_name': 'resource-1'},
                {'display_name': 'resource-2'},
                {'display_name': 'resource-3'}
            ]
        }
        mocked_resource = self.get_mocked_resource(response=resp_resources,
                                                   response_repeat=3)

        self.assertEqual([{'display_name': 'resource-1'}],
                         mocked_resource.find_by_display_name('resource-1'))
        self.assertEqual([{'display_name': 'resource-2'}],
                         mocked_resource.find_by_display_name('resource-2'))
        self.assertEqual([{'display_name': 'resource-3'}],
                         mocked_resource.find_by_display_name('resource-3'))

        resp_resources = {
            'results': [
                {'display_name': 'resource-1'},
                {'display_name': 'resource-1'},
                {'display_name': 'resource-1'}
            ]
        }
        mocked_resource = self.get_mocked_resource(response=resp_resources)
        self.assertEqual(resp_resources['results'],
                         mocked_resource.find_by_display_name('resource-1'))

    def test_list_all(self):
        mocked_resource = self.get_mocked_resource()
        mocked_resource.list()
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/switching-profiles/'
            '?include_system_owned=True',
            data=None,
            headers=self.default_headers())


class LogicalPortTestCase(BaseTestResource):

    def setUp(self):
        super(LogicalPortTestCase, self).setUp(resources.LogicalPort)

    def _get_profile_dicts(self, fake_port):
        fake_profile_dicts = []
        for profile_id in fake_port['switching_profile_ids']:
            fake_profile_dicts.append({'resource_type': profile_id['key'],
                                       'id': profile_id['value']})
        return fake_profile_dicts

    def _get_pktcls_bindings(self):
        fake_pkt_classifiers = []
        fake_binding_repr = []
        for i in range(0, 3):
            ip = "9.10.11.%s" % i
            mac = "00:0c:29:35:4a:%sc" % i
            fake_pkt_classifiers.append(resources.PacketAddressClassifier(
                ip, mac, None))
            fake_binding_repr.append({
                'ip_address': ip,
                'mac_address': mac
            })
        return fake_pkt_classifiers, fake_binding_repr

    def test_create_logical_port(self):
        """Test creating a port.

        returns the correct response and 200 status
        """
        fake_port = test_constants.FAKE_PORT.copy()

        profile_dicts = self._get_profile_dicts(fake_port)

        pkt_classifiers, binding_repr = self._get_pktcls_bindings()

        mocked_resource = self.get_mocked_resource()
        description = 'dummy'
        switch_profile = resources.SwitchingProfile
        mocked_resource.create(
            fake_port['logical_switch_id'],
            fake_port['attachment']['id'],
            address_bindings=pkt_classifiers,
            switch_profile_ids=switch_profile.build_switch_profile_ids(
                mock.Mock(), *profile_dicts),
            description=description)

        resp_body = {
            'logical_switch_id': fake_port['logical_switch_id'],
            'switching_profile_ids': fake_port['switching_profile_ids'],
            'attachment': {
                'attachment_type': 'VIF',
                'id': fake_port['attachment']['id']
            },
            'admin_state': 'UP',
            'address_bindings': binding_repr,
            'description': description
        }

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/logical-ports',
            data=jsonutils.dumps(resp_body, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_port_with_attachtype_cif(self):
        """Test creating a port returns the correct response and 200 status

        """
        fake_port = copy.deepcopy(test_constants.FAKE_CONTAINER_PORT)

        profile_dicts = self._get_profile_dicts(fake_port)

        pkt_classifiers, binding_repr = self._get_pktcls_bindings()

        fake_port['address_bindings'] = binding_repr

        mocked_resource = self.get_mocked_resource()
        switch_profile = resources.SwitchingProfile
        fake_port_ctx = fake_port['attachment']['context']

        fake_container_host_vif_id = fake_port_ctx['container_host_vif_id']

        mocked_resource.create(
            fake_port['logical_switch_id'],
            fake_port['attachment']['id'],
            parent_vif_id=fake_container_host_vif_id,
            traffic_tag=fake_port_ctx['vlan_tag'],
            address_bindings=pkt_classifiers,
            switch_profile_ids=switch_profile.build_switch_profile_ids(
                mock.Mock(), *profile_dicts),
            vif_type=fake_port_ctx['vif_type'], app_id=fake_port_ctx['app_id'],
            allocate_addresses=fake_port_ctx['allocate_addresses'])

        resp_body = {
            'logical_switch_id': fake_port['logical_switch_id'],
            'switching_profile_ids': fake_port['switching_profile_ids'],
            'attachment': {
                'attachment_type': 'VIF',
                'id': fake_port['attachment']['id'],
                'context': {
                    'resource_type': 'VifAttachmentContext',
                    'allocate_addresses': 'Both',
                    'parent_vif_id': fake_container_host_vif_id,
                    'traffic_tag': fake_port_ctx['vlan_tag'],
                    'app_id': fake_port_ctx['app_id'],
                    'vif_type': 'CHILD',
                }
            },
            'admin_state': 'UP',
            'address_bindings': fake_port['address_bindings']
        }

        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/logical-ports',
            data=jsonutils.dumps(resp_body, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_port_admin_down(self):
        """Test creating port with admin_state down."""
        fake_port = test_constants.FAKE_PORT
        fake_port['admin_state'] = "DOWN"
        mocked_resource = self.get_mocked_resource(response=fake_port)

        result = mocked_resource.create(
            test_constants.FAKE_PORT['logical_switch_id'],
            test_constants.FAKE_PORT['attachment']['id'],
            tags={}, admin_state=False)

        self.assertEqual(fake_port, result)

    def test_delete_resource(self):
        """Test deleting port."""
        super(LogicalPortTestCase, self).test_delete_resource(
            extra_params='detach=true')

    def test_get_logical_port_by_attachment(self):
        """Test deleting port."""
        mocked_resource = self.get_mocked_resource()
        attachment_type = nsx_constants.ATTACHMENT_DHCP
        attachment_id = '1234'
        mocked_resource.get_by_attachment(attachment_type, attachment_id)
        test_client.assert_json_call(
            'get', mocked_resource,
            "https://1.2.3.4/api/v1/logical-ports/?attachment_type=%s"
            "&attachment_id=%s" % (attachment_type, attachment_id),
            headers=self.default_headers())

    def test_clear_port_bindings(self):
        fake_port = copy.copy(test_constants.FAKE_PORT)
        fake_port['address_bindings'] = ['a', 'b']
        mocked_resource = self.get_mocked_resource()

        def get_fake_port(*args, **kwargs):
            return copy.copy(fake_port)

        mocked_resource.client.get = get_fake_port
        mocked_resource.update(
            fake_port['id'], fake_port['attachment']['id'],
            address_bindings=[])

        fake_port['address_bindings'] = []
        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/logical-ports/%s' % fake_port['id'],
            data=jsonutils.dumps(fake_port, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_port_fail(self):
        """Test the failure of port creation."""
        fake_port = test_constants.FAKE_PORT.copy()
        profile_dicts = self._get_profile_dicts(fake_port)
        pkt_classifiers, binding_repr = self._get_pktcls_bindings()
        fake_port['address_bindings'] = binding_repr
        mocked_resource = self.get_mocked_resource(mock_validate=False)
        switch_profile = resources.SwitchingProfile
        try:
            mocked_resource.create(
                fake_port['logical_switch_id'],
                fake_port['attachment']['id'],
                address_bindings=pkt_classifiers,
                switch_profile_ids=switch_profile.build_switch_profile_ids(
                    mock.Mock(), *profile_dicts))
        except exceptions.ManagerError as e:
            self.assertIn(nsxlib_testcase.NSX_MANAGER, e.msg)

    def test_update_logical_port_no_addr_binding(self):
        fake_port = copy.deepcopy(test_constants.FAKE_CONTAINER_PORT)
        mocked_resource = self.get_mocked_resource()
        new_name = 'updated_port'
        new_desc = 'updated'
        fake_port_ctx = fake_port['attachment']['context']
        fake_container_host_vif_id = fake_port_ctx['container_host_vif_id']

        def get_fake_port(*args, **kwargs):
            return copy.copy(fake_port)

        mocked_resource.client.get = get_fake_port

        mocked_resource.update(
            fake_port['id'],
            fake_port['attachment']['id'],
            name=new_name,
            description=new_desc,
            parent_vif_id=fake_container_host_vif_id,
            traffic_tag=fake_port_ctx['vlan_tag'],
            vif_type=fake_port_ctx['vif_type'],
            app_id=fake_port_ctx['app_id'],
            allocate_addresses=fake_port_ctx['allocate_addresses'])

        fake_port['display_name'] = new_name
        fake_port['description'] = new_desc
        fake_port['attachment'] = {
            'attachment_type': 'VIF',
            'id': fake_port['attachment']['id'],
            'context': {
                'resource_type': 'VifAttachmentContext',
                'allocate_addresses': 'Both',
                'parent_vif_id': fake_container_host_vif_id,
                'traffic_tag': fake_port_ctx['vlan_tag'],
                'app_id': fake_port_ctx['app_id'],
                'vif_type': 'CHILD',
            }
        }

        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/logical-ports/%s' % fake_port['id'],
            data=jsonutils.dumps(fake_port, sort_keys=True),
            headers=self.default_headers())

    def test_update_logical_port_with_addr_binding(self):
        fake_port = copy.deepcopy(test_constants.FAKE_CONTAINER_PORT)
        mocked_resource = self.get_mocked_resource()
        new_name = 'updated_port'
        new_desc = 'updated'
        fake_port_ctx = fake_port['attachment']['context']
        fake_container_host_vif_id = fake_port_ctx['container_host_vif_id']
        pkt_classifiers, binding_repr = self._get_pktcls_bindings()

        def get_fake_port(*args, **kwargs):
            return copy.copy(fake_port)

        mocked_resource.client.get = get_fake_port

        mocked_resource.update(
            fake_port['id'],
            fake_port['attachment']['id'],
            name=new_name,
            description=new_desc,
            parent_vif_id=fake_container_host_vif_id,
            traffic_tag=fake_port_ctx['vlan_tag'],
            vif_type=fake_port_ctx['vif_type'],
            app_id=fake_port_ctx['app_id'],
            allocate_addresses=fake_port_ctx['allocate_addresses'],
            address_bindings=pkt_classifiers)

        fake_port['display_name'] = new_name
        fake_port['description'] = new_desc
        fake_port['attachment'] = {
            'attachment_type': 'VIF',
            'id': fake_port['attachment']['id'],
            'context': {
                'resource_type': 'VifAttachmentContext',
                'allocate_addresses': 'Both',
                'parent_vif_id': fake_container_host_vif_id,
                'traffic_tag': fake_port_ctx['vlan_tag'],
                'app_id': fake_port_ctx['app_id'],
                'vif_type': 'CHILD',
            }
        }
        fake_port['address_bindings'] = binding_repr
        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/logical-ports/%s' % fake_port['id'],
            data=jsonutils.dumps(fake_port, sort_keys=True),
            headers=self.default_headers())


class LogicalRouterTestCase(BaseTestResource):

    def setUp(self):
        super(LogicalRouterTestCase, self).setUp(
            core_resources.NsxLibLogicalRouter)

    def test_create_logical_router(self):
        """Test creating a router returns the correct response and 201 status.

        """
        fake_router = test_constants.FAKE_ROUTER.copy()
        router = self.get_mocked_resource()
        tier0_router = True
        description = 'dummy'
        router.create(fake_router['display_name'], None, None, tier0_router,
                      description=description)

        data = {
            'display_name': fake_router['display_name'],
            'router_type': 'TIER0' if tier0_router else 'TIER1',
            'tags': None,
            'description': description
        }

        test_client.assert_json_call(
            'post', router,
            'https://1.2.3.4/api/v1/logical-routers',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_force_delete_logical_router(self):
        """Test force deleting router"""
        router = self.get_mocked_resource()
        uuid = test_constants.FAKE_ROUTER['id']
        router.delete(uuid, True)
        test_client.assert_json_call(
            'delete', router,
            'https://1.2.3.4/api/v1/logical-routers/%s?force=True' % uuid,
            headers=self.default_headers())

    def test_list_logical_router_by_type(self):
        router = self.get_mocked_resource()
        router_type = 'TIER0'
        router.list(router_type=router_type)
        test_client.assert_json_call(
            'get', router,
            'https://1.2.3.4/api/v1/logical-routers?router_type=%s' %
            router_type)

    def test_get_logical_router_fw_section(self):
        fake_router = test_constants.FAKE_ROUTER.copy()
        router = self.get_mocked_resource()
        section_id = router.get_firewall_section_id(
            test_constants.FAKE_ROUTER_UUID, router_body=fake_router)
        self.assertEqual(test_constants.FAKE_ROUTER_FW_SEC_UUID, section_id)

    def _test_nat_rule_create(self, nsx_version, add_bypas_arg):
        router = self.get_mocked_resource()
        action = 'SNAT'
        translated_net = '1.1.1.1'
        priority = 10

        data = {
            'action': action,
            'enabled': True,
            'translated_network': translated_net,
            'rule_priority': priority
        }
        if add_bypas_arg:
            # Expect nat_pass to be sent to the backend
            data['nat_pass'] = False

        # Ignoring 'bypass_firewall' with version 1.1
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value=nsx_version):
            router.add_nat_rule(test_constants.FAKE_ROUTER_UUID,
                                action=action,
                                translated_network=translated_net,
                                rule_priority=priority,
                                bypass_firewall=False)
            test_client.assert_json_call(
                'post', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/nat/rules' %
                    test_constants.FAKE_ROUTER_UUID),
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_nat_rule_create_v1(self):
        # Ignoring 'bypass_firewall' with version 1.1
        self._test_nat_rule_create('1.1.0', False)

    def test_nat_rule_create_v2(self):
        # Sending 'bypass_firewall' with version 1.1
        self._test_nat_rule_create('2.0.0', True)

    def test_nat_rule_list(self):
        router = self.get_mocked_resource()
        router.list_nat_rules(test_constants.FAKE_ROUTER_UUID)
        test_client.assert_json_call(
            'get', router,
            ('https://1.2.3.4/api/v1/logical-routers/%s/nat/rules' %
                test_constants.FAKE_ROUTER_UUID),
            headers=self.default_headers())

    def test_nat_rule_update(self):
        router = self.get_mocked_resource()
        rule_id = '123'
        with mock.patch.object(router.client, 'get',
                               return_value={'id': rule_id}):
            router.update_nat_rule(test_constants.FAKE_ROUTER_UUID,
                                   rule_id, nat_pass=False)
            data = {'id': rule_id, 'nat_pass': False}
            test_client.assert_json_call(
                'put', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/nat/rules/%s' %
                    (test_constants.FAKE_ROUTER_UUID, rule_id)),
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_delete_nat_rule_by_gw(self):
        router = self.get_mocked_resource()
        rule_id = '123'
        router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '3.3.3.3'
        existing_rules = [{
            'translated_network': gw_ip,
            'logical_router_id': router_id,
            'id': rule_id,
            'action': 'SNAT',
            'resource_type': 'NatRule'}]
        with mock.patch.object(router.client, 'list',
                               return_value={'results': existing_rules}):
            router.delete_nat_rule_by_values(router_id,
                                             translated_network=gw_ip)
            test_client.assert_json_call(
                'delete', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/nat/rules/%s' %
                    (router_id, rule_id)),
                headers=self.default_headers())

    def test_delete_nat_rule_by_gw_and_source(self):
        router = self.get_mocked_resource()
        rule_id = '123'
        router_id = test_constants.FAKE_ROUTER_UUID
        gw_ip = '3.3.3.3'
        source_net = '4.4.4.4'
        existing_rules = [{
            'translated_network': gw_ip,
            'logical_router_id': router_id,
            'id': rule_id,
            'match_source_network': source_net,
            'action': 'SNAT',
            'resource_type': 'NatRule'}]
        with mock.patch.object(router.client, 'list',
                               return_value={'results': existing_rules}):
            router.delete_nat_rule_by_values(router_id,
                                             translated_network=gw_ip,
                                             match_source_network=source_net)
            test_client.assert_json_call(
                'delete', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/nat/rules/%s' %
                    (router_id, rule_id)),
                headers=self.default_headers())

    def test_update_advertisement(self):
        router = self.get_mocked_resource()
        router_id = test_constants.FAKE_ROUTER_UUID
        data = {'advertise_nat_routes': 'a',
                'advertise_nsx_connected_routes': 'b',
                'advertise_static_routes': False,
                'enabled': True,
                'advertise_lb_vip': False,
                'advertise_lb_snat_ip': False}
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.1.0'), \
            mock.patch.object(router.client, 'get',
                              return_value={}):
            router.update_advertisement(
                router_id, **data)
            test_client.assert_json_call(
                'put', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/routing/'
                 'advertisement' % router_id),
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_update_advertisement_no_lb(self):
        router = self.get_mocked_resource()
        router_id = test_constants.FAKE_ROUTER_UUID
        data = {'advertise_nat_routes': 'a',
                'advertise_nsx_connected_routes': 'b',
                'advertise_static_routes': False,
                'enabled': True}
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='1.1.0'), \
            mock.patch.object(router.client, 'get',
                              return_value={}):
            # lb args will be ignored on this nsx version
            router.update_advertisement(
                router_id,
                advertise_lb_vip=False,
                advertise_lb_snat_ip=False,
                **data)
            test_client.assert_json_call(
                'put', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/routing/'
                 'advertisement' % router_id),
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_update_advertisement_rules(self):
        router = self.get_mocked_resource()
        router_id = test_constants.FAKE_ROUTER_UUID
        rules = [{"action": "ALLOW",
                  "networks": ["44.0.0.0/20"],
                  "display_name": "rule1"},
                 {"action": "ALLOW",
                  "networks": ["6.60.0.0/20"],
                  "display_name": "rule2"}]
        with mock.patch.object(router.client, 'get',
                               return_value={}):
            router.update_advertisement_rules(router_id, rules)
            test_client.assert_json_call(
                'put', router,
                ('https://1.2.3.4/api/v1/logical-routers/%s/routing/'
                 'advertisement/rules' % router_id),
                data=jsonutils.dumps({'rules': rules}, sort_keys=True),
                headers=self.default_headers())

    def test_get_advertisement_rules(self):
        router = self.get_mocked_resource()
        router_id = test_constants.FAKE_ROUTER_UUID
        router.get_advertisement_rules(router_id)
        test_client.assert_json_call(
            'get', router,
            ('https://1.2.3.4/api/v1/logical-routers/%s/routing/'
             'advertisement/rules' % router_id),
            headers=self.default_headers())


class LogicalRouterPortTestCase(BaseTestResource):

    def setUp(self):
        super(LogicalRouterPortTestCase, self).setUp(
            resources.LogicalRouterPort)

    def test_create_logical_router_port(self):
        """Test creating a router port.

        returns the correct response and 201 status
        """
        fake_router_port = test_constants.FAKE_ROUTER_PORT.copy()
        fake_relay_uuid = uuidutils.generate_uuid()
        lrport = self.get_mocked_resource()

        data = {
            'display_name': fake_router_port['display_name'],
            'logical_router_id': fake_router_port['logical_router_id'],
            'resource_type': fake_router_port['resource_type'],
            'tags': [],
            'service_bindings': [{'service_id': {
                'target_type': 'LogicalService',
                'target_id': fake_relay_uuid}}]
        }

        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.0.0'):
            lrport.create(fake_router_port['logical_router_id'],
                          fake_router_port['display_name'],
                          None,
                          fake_router_port['resource_type'],
                          None, None, None,
                          relay_service_uuid=fake_relay_uuid)

            test_client.assert_json_call(
                'post', lrport,
                'https://1.2.3.4/api/v1/logical-router-ports',
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_logical_router_port_max_attempts(self):
        """Test a router port api has the configured retries."""
        lrport = self.get_mocked_resource()

        self.assertEqual(nsxlib_testcase.NSX_MAX_ATTEMPTS,
                         lrport.client.max_attempts)

    def test_update_logical_router_port(self):
        fake_router_port = test_constants.FAKE_ROUTER_PORT.copy()
        uuid = fake_router_port['id']
        fake_relay_uuid = uuidutils.generate_uuid()
        lrport = self.get_mocked_resource()
        with mock.patch.object(lrport.client, 'get',
                               return_value=fake_router_port),\
            mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                       return_value='2.0.0'):
            lrport.update(uuid, relay_service_uuid=fake_relay_uuid)
            data = {
                'id': uuid,
                'display_name': fake_router_port['display_name'],
                'logical_router_id': fake_router_port['logical_router_id'],
                'resource_type': fake_router_port['resource_type'],
                "revision": 0,
                'service_bindings': [{'service_id': {
                    'target_type': 'LogicalService',
                    'target_id': fake_relay_uuid}}]
            }

            test_client.assert_json_call(
                'put', lrport,
                'https://1.2.3.4/api/v1/logical-router-ports/%s' % uuid,
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_get_logical_router_port_by_router_id(self):
        """Test getting a router port by router id."""
        fake_router_port = test_constants.FAKE_ROUTER_PORT.copy()
        resp_resources = {'results': [fake_router_port]}
        lrport = self.get_mocked_resource(response=resp_resources)

        router_id = fake_router_port['logical_router_id']
        result = lrport.get_by_router_id(router_id)
        self.assertEqual(fake_router_port, result[0])
        test_client.assert_json_call(
            'get', lrport,
            'https://1.2.3.4/api/v1/logical-router-ports/?'
            'logical_router_id=%s' % router_id,
            headers=self.default_headers())

    def test_get_logical_router_port_by_switch_id(self):
        """Test getting a router port by switch id."""
        fake_router_port = test_constants.FAKE_ROUTER_PORT.copy()
        resp_resources = {
            'result_count': 1,
            'results': [fake_router_port]
        }
        lrport = self.get_mocked_resource(response=resp_resources)

        switch_id = test_constants.FAKE_SWITCH_UUID
        lrport.get_by_lswitch_id(switch_id)
        test_client.assert_json_call(
            'get', lrport,
            'https://1.2.3.4/api/v1/logical-router-ports/?'
            'logical_switch_id=%s' % switch_id,
            headers=self.default_headers())


class IpPoolTestCase(BaseTestResource):

    def setUp(self):
        super(IpPoolTestCase, self).setUp(resources.IpPool)

    def test_create_ip_pool_all_args(self):
        """Test creating an IP pool

        returns the correct response and 201 status
        """
        pool = self.get_mocked_resource()

        display_name = 'dummy'
        gateway_ip = '1.1.1.1'
        ranges = [{'start': '2.2.2.0', 'end': '2.2.2.255'},
                  {'start': '3.2.2.0', 'end': '3.2.2.255'}]
        cidr = '2.2.2.0/24'
        description = 'desc'
        dns_nameserver = '7.7.7.7'
        pool.create(cidr, allocation_ranges=ranges,
                    display_name=display_name,
                    gateway_ip=gateway_ip,
                    description=description,
                    dns_nameservers=[dns_nameserver])

        data = {
            'display_name': display_name,
            'description': description,
            'subnets': [{
                'gateway_ip': gateway_ip,
                'allocation_ranges': ranges,
                'cidr': cidr,
                'dns_nameservers': [dns_nameserver]
            }]
        }

        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_ip_pool_minimal_args(self):
        pool = self.get_mocked_resource()

        ranges = [{'start': '2.2.2.0', 'end': '2.2.2.255'},
                  {'start': '3.2.2.0', 'end': '3.2.2.255'}]
        cidr = '2.2.2.0/24'
        pool.create(cidr, allocation_ranges=ranges)

        data = {
            'subnets': [{
                'allocation_ranges': ranges,
                'cidr': cidr,
            }]
        }

        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_ip_pool_no_ranges_with_gateway(self):
        pool = self.get_mocked_resource()
        cidr = '2.2.2.0/30'
        gateway_ip = '2.2.2.1'
        pool.create(cidr, allocation_ranges=None, gateway_ip=gateway_ip)
        exp_ranges = [{'start': '2.2.2.0', 'end': '2.2.2.0'},
                      {'start': '2.2.2.2', 'end': '2.2.2.3'}]

        data = {
            'subnets': [{
                'gateway_ip': gateway_ip,
                'allocation_ranges': exp_ranges,
                'cidr': cidr,
            }]
        }

        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_ip_pool_no_ranges_no_gateway(self):
        pool = self.get_mocked_resource()
        cidr = '2.2.2.0/30'
        pool.create(cidr, allocation_ranges=None)
        exp_ranges = [{'start': '2.2.2.0', 'end': '2.2.2.3'}]

        data = {
            'subnets': [{
                'allocation_ranges': exp_ranges,
                'cidr': cidr,
            }]
        }

        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_ip_pool_no_cidr(self):
        pool = self.get_mocked_resource()
        gateway_ip = '1.1.1.1'
        ranges = [{'start': '2.2.2.0', 'end': '2.2.2.255'},
                  {'start': '3.2.2.0', 'end': '3.2.2.255'}]
        cidr = None

        try:
            pool.create(cidr, allocation_ranges=ranges,
                        gateway_ip=gateway_ip)
        except exceptions.InvalidInput:
            # This call should fail
            pass
        else:
            self.fail("shouldn't happen")

    def test_update_ip_pool_name(self):
        fake_ip_pool = test_constants.FAKE_IP_POOL.copy()
        pool = self.get_mocked_resource(response=fake_ip_pool)

        uuid = fake_ip_pool['id']
        new_name = 'new_name'
        pool.update(uuid, display_name=new_name)
        fake_ip_pool['display_name'] = new_name
        test_client.assert_json_call(
            'put', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s' % uuid,
            data=jsonutils.dumps(fake_ip_pool, sort_keys=True),
            headers=self.default_headers())

    def test_update_ip_pool_gateway(self):
        fake_ip_pool = test_constants.FAKE_IP_POOL.copy()
        pool = self.get_mocked_resource(response=fake_ip_pool)

        uuid = fake_ip_pool['id']
        new_gateway = '1.0.0.1'
        pool.update(uuid, gateway_ip=new_gateway)
        fake_ip_pool["subnets"][0]['gateway_ip'] = new_gateway
        test_client.assert_json_call(
            'put', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s' % uuid,
            data=jsonutils.dumps(fake_ip_pool, sort_keys=True),
            headers=self.default_headers())

    def test_update_ip_pool_delete_gateway(self):
        fake_ip_pool = test_constants.FAKE_IP_POOL.copy()
        pool = self.get_mocked_resource(response=fake_ip_pool)

        uuid = fake_ip_pool['id']
        pool.update(uuid, gateway_ip=None)
        del fake_ip_pool["subnets"][0]['gateway_ip']
        test_client.assert_json_call(
            'put', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s' % uuid,
            data=jsonutils.dumps(fake_ip_pool, sort_keys=True),
            headers=self.default_headers())

    def test_allocate_ip_from_pool(self):
        pool = self.get_mocked_resource()

        uuid = test_constants.FAKE_IP_POOL['id']
        addr = '1.1.1.1'
        pool.allocate(uuid, ip_addr=addr)

        data = {'allocation_id': addr}
        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s?action=ALLOCATE' % uuid,
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_release_ip_to_pool(self):
        pool = self.get_mocked_resource()

        uuid = test_constants.FAKE_IP_POOL['id']
        addr = '1.1.1.1'
        pool.release(uuid, addr)

        data = {'allocation_id': addr}
        test_client.assert_json_call(
            'post', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s?action=RELEASE' % uuid,
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_get_ip_pool_allocations(self):
        """Test getting a router port by router id"""
        fake_ip_pool = test_constants.FAKE_IP_POOL.copy()
        pool = self.get_mocked_resource(response=fake_ip_pool)

        uuid = fake_ip_pool['id']
        result = pool.get_allocations(uuid)
        self.assertEqual(fake_ip_pool, result)
        test_client.assert_json_call(
            'get', pool,
            'https://1.2.3.4/api/v1/pools/ip-pools/%s/allocations' % uuid,
            headers=self.default_headers())


class TestNsxSearch(nsxlib_testcase.NsxClientTestCase):

    def test_nsx_search_tags(self):
        """Test search of resources with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'scope': 'user', 'tag': 'k8s'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags)
            search.assert_called_with('search?query=%s' % query)

    def test_nsx_search_tags_scope_only(self):
        """Test search of resources with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'scope': 'user'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags)
            search.assert_called_with('search?query=%s' % query)

    def test_nsx_search_tags_tag_only(self):
        """Test search of resources with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'tag': 'k8s'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags)
            search.assert_called_with('search?query=%s' % query)

    def test_nsx_search_tags_tag_and_scope(self):
        """Test search of resources with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'tag': 'k8s'}, {'scope': 'user'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags)
            search.assert_called_with('search?query=%s' % query)

    def test_nsx_search_tags_and_resource_type(self):
        """Test search of specified resource with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'scope': 'user', 'tag': 'k8s'}]
            res_type = 'LogicalPort'
            query = self.nsxlib._build_query(tags=user_tags)
            # Add resource_type to the query
            query = "resource_type:%s AND %s" % (res_type, query)
            self.nsxlib.search_by_tags(tags=user_tags, resource_type=res_type)
            search.assert_called_with('search?query=%s' % query)

    def test_nsx_search_tags_and_cursor(self):
        """Test search of resources with the specified tag and cursor."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'scope': 'user', 'tag': 'k8s'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags, cursor=50)
            search.assert_called_with('search?query=%s&cursor=50' % query)

    def test_nsx_search_tags_and_page_size(self):
        """Test search of resources with the specified tag and page size."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            user_tags = [{'scope': 'user', 'tag': 'k8s'}]
            query = self.nsxlib._build_query(tags=user_tags)
            self.nsxlib.search_by_tags(tags=user_tags, page_size=100)
            search.assert_called_with('search?query=%s&page_size=100' % query)

    def test_nsx_search_invalid_query_fail(self):
        """Test search query failure for missing tag argument."""
        self.assertRaises(exceptions.NsxSearchInvalidQuery,
                          self.nsxlib.search_by_tags,
                          tags=None, resource_type=None)

    def test_nsx_search_invalid_tags_fail(self):
        """Test search of resources with the invalid tag."""
        user_tags = [{'scope': 'user', 'invalid_tag_key': 'k8s'}]
        self.assertRaises(exceptions.NsxSearchInvalidQuery,
                          self.nsxlib._build_query,
                          tags=user_tags)

    def test_nsx_search_all_by_tags(self):
        """Test search all of resources with the specified tag."""
        with mock.patch.object(self.nsxlib.client, 'url_get') as search:
            search.side_effect = [
                {"cursor": "2",
                 "result_count": 3,
                 "results": [{"id": "s1"},
                             {"id": "s2"}]},
                {"cursor": "3",
                 "result_count": 3,
                 "results": [{"id": "s3"}]}]
            user_tags = [{'scope': 'user', 'tag': 'k8s'}]
            query = self.nsxlib._build_query(tags=user_tags)
            results = self.nsxlib.search_all_by_tags(tags=user_tags)
            search.assert_has_calls([
                mock.call('search?query=%s' % query),
                mock.call('search?query=%s&cursor=2' % query)])
            self.assertEqual(3, len(results))

    def test_get_id_by_resource_and_tag(self):
        id = 'test'
        scope = 'user'
        tag = 'k8s'
        res_type = 'LogicalPort'
        results = {'result_count': 1, 'results': [{'id': id}]}
        with mock.patch.object(self.nsxlib.client, 'url_get',
                               return_value=results):
            actual_id = self.nsxlib.get_id_by_resource_and_tag(
                res_type, scope, tag)
            self.assertEqual(id, actual_id)

    def test_get_id_by_resource_and_tag_not_found(self):
        scope = 'user'
        tag = 'k8s'
        res_type = 'LogicalPort'
        results = {'result_count': 0, 'results': []}
        with mock.patch.object(self.nsxlib.client, 'url_get',
                               return_value=results):
            self.assertRaises(exceptions.ResourceNotFound,
                              self.nsxlib.get_id_by_resource_and_tag,
                              res_type, scope, tag, alert_not_found=True)

    def test_get_id_by_resource_and_tag_multiple(self):
        scope = 'user'
        tag = 'k8s'
        res_type = 'LogicalPort'
        results = {'result_count': 2, 'results': [{'id': '1'}, {'id': '2'}]}
        with mock.patch.object(self.nsxlib.client, 'url_get',
                               return_value=results):
            self.assertRaises(exceptions.ManagerError,
                              self.nsxlib.get_id_by_resource_and_tag,
                              res_type, scope, tag, alert_multiple=True)


class TransportZone(BaseTestResource):

    def setUp(self):
        super(TransportZone, self).setUp(core_resources.NsxLibTransportZone)

    def test_get_transport_zone_type(self):
        fake_tz = test_constants.FAKE_TZ.copy()
        tz = self.get_mocked_resource()
        with mock.patch.object(tz.client, 'url_get', return_value=fake_tz):
            tz_type = tz.get_transport_type(fake_tz['id'])
            self.assertEqual(tz.TRANSPORT_TYPE_OVERLAY, tz_type)

            # call it again to test it when cached
            tz_type = tz.get_transport_type(fake_tz['id'])
            self.assertEqual(tz.TRANSPORT_TYPE_OVERLAY, tz_type)

    def test_get_host_switch_mode(self):
        fake_tz = test_constants.FAKE_TZ.copy()
        tz = self.get_mocked_resource()
        with mock.patch.object(tz.client, 'url_get', return_value=fake_tz):
            tz_mode = tz.get_host_switch_mode(fake_tz['id'])
            self.assertEqual(tz.HOST_SWITCH_MODE_STANDARD, tz_mode)


class MetadataProxy(BaseTestResource):

    def setUp(self):
        super(MetadataProxy, self).setUp(core_resources.NsxLibMetadataProxy)

    def test_update_metadata_proxy(self):
        fake_md = test_constants.FAKE_MD.copy()
        md = self.get_mocked_resource()
        new_url = "http://2.2.2.20:3500/xyz"
        new_secret = 'abc'
        new_edge = uuidutils.generate_uuid()
        with mock.patch.object(md.client, 'url_get', return_value=fake_md):
            md.update(fake_md['id'], server_url=new_url, secret=new_secret,
                      edge_cluster_id=new_edge)
            fake_md.update({'metadata_server_url': new_url,
                            'secret': new_secret,
                            'edge_cluster_id': new_edge})
            test_client.assert_json_call(
                'put', md,
                'https://1.2.3.4/api/v1/md-proxies/%s' % fake_md['id'],
                data=jsonutils.dumps(fake_md, sort_keys=True),
                headers=self.default_headers())


class NsxLibSwitchTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibSwitchTestCase, self).setUp(
            core_resources.NsxLibLogicalSwitch)
        self._tz_id = uuidutils.generate_uuid()

    def _create_body(self, admin_state=nsx_constants.ADMIN_STATE_UP,
                     vlan_id=None, description=None, trunk_vlan=None):
        body = {
            "transport_zone_id": self._tz_id,
            "replication_mode": "MTEP",
            "display_name": "fake_name",
            "tags": [],
            "admin_state": admin_state
        }
        if vlan_id:
            body['vlan'] = vlan_id
        if description is not None:
            body['description'] = description
        if trunk_vlan:
            body['vlan_trunk_spec'] = {
                'vlan_ranges': [{'start': trunk_vlan[0],
                                 'end': trunk_vlan[1]}]}
        return body

    def test_create_logical_switch(self):
        """Test creating a switch returns the correct response and 200 status

        """
        desc = 'dummy'
        ls = self.get_mocked_resource()
        ls.create(mocks.FAKE_NAME, self._tz_id, [],
                  description=desc)
        data = self._create_body(description=desc)
        test_client.assert_json_call(
            'post', ls,
            'https://1.2.3.4/api/v1/logical-switches',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_switch_admin_down(self):
        """Test creating switch with admin_state down"""
        ls = self.get_mocked_resource()
        ls.create(mocks.FAKE_NAME, self._tz_id, [],
                  admin_state=False)
        data = self._create_body(admin_state=nsx_constants.ADMIN_STATE_DOWN)
        test_client.assert_json_call(
            'post', ls,
            'https://1.2.3.4/api/v1/logical-switches',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_switch_vlan(self):
        """Test creating switch with provider:network_type VLAN"""
        ls = self.get_mocked_resource()
        vlan_id = '123'
        ls.create(mocks.FAKE_NAME, self._tz_id, [],
                  vlan_id=vlan_id)
        data = self._create_body(vlan_id=vlan_id)
        test_client.assert_json_call(
            'post', ls,
            'https://1.2.3.4/api/v1/logical-switches',
            data=jsonutils.dumps(data, sort_keys=True),
            headers=self.default_headers())

    def test_create_logical_switch_trunk(self):
        """Test creating switch with trunk vlan"""
        ls = self.get_mocked_resource()
        trunk_vlan = [10, 20]
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.2.0'):
            ls.create(mocks.FAKE_NAME, self._tz_id, [],
                      trunk_vlan_range=trunk_vlan)
            data = self._create_body(trunk_vlan=trunk_vlan)
            test_client.assert_json_call(
                'post', ls,
                'https://1.2.3.4/api/v1/logical-switches',
                data=jsonutils.dumps(data, sort_keys=True),
                headers=self.default_headers())

    def test_create_logical_switch_trunk_not_supported(self):
        """Test creating switch with trunk vlan without the support"""
        ls = self.get_mocked_resource()
        trunk_vlan = [10, 20]
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.0.0'):
            self.assertRaises(exceptions.InvalidInput,
                              ls.create,
                              mocks.FAKE_NAME, self._tz_id, [],
                              trunk_vlan_range=trunk_vlan)

    def test_create_logical_switch_trunk_with_vlan(self):
        """Test creating switch with trunk vlan and vlan tag"""
        ls = self.get_mocked_resource()
        trunk_vlan = [10, 20]
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.2.0'):
            self.assertRaises(exceptions.InvalidInput,
                              ls.create,
                              mocks.FAKE_NAME, self._tz_id, [],
                              trunk_vlan_range=trunk_vlan,
                              vlan_id='111')

    def test_create_logical_switch_illegal_trunk(self):
        """Test creating switch with illegal trunk vlan"""
        ls = self.get_mocked_resource()
        trunk_vlan = [10]
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.2.0'):
            self.assertRaises(exceptions.InvalidInput,
                              ls.create,
                              mocks.FAKE_NAME, self._tz_id, [],
                              trunk_vlan_range=trunk_vlan)

    def test_delete_resource(self):
        """Test deleting switch"""
        super(NsxLibSwitchTestCase, self).test_delete_resource(
            extra_params='detach=true&cascade=true')


class NsxLibPortMirrorTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibPortMirrorTestCase, self).setUp(
            core_resources.NsxLibPortMirror)


class NsxLibBridgeEndpointTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibBridgeEndpointTestCase, self).setUp(
            core_resources.NsxLibBridgeEndpoint)


class NsxLibEdgeClusterTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibEdgeClusterTestCase, self).setUp(
            core_resources.NsxLibEdgeCluster)


class NsxLibDhcpProfileTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibDhcpProfileTestCase, self).setUp(
            core_resources.NsxLibDhcpProfile)


class NsxLibDhcpRelayServiceTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibDhcpRelayServiceTestCase, self).setUp(
            core_resources.NsxLibDhcpRelayService)

    def test_server_ips(self):
        fake_srv = test_constants.FAKE_RELAY_SERVICE.copy()
        relay_service = self.get_mocked_resource()
        with mock.patch.object(relay_service.client, 'url_get',
                               return_value=fake_srv), \
            mock.patch.object(self.nsxlib.client, 'url_get',
                              return_value=test_constants.FAKE_RELAY_PROFILE):
            server_ips = relay_service.get_server_ips(fake_srv['id'])
            self.assertEqual(1, len(server_ips))
            self.assertEqual(test_constants.FAKE_RELAY_SERVER,
                             server_ips[0])


class NsxLibDhcpRelayProfileTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibDhcpRelayProfileTestCase, self).setUp(
            core_resources.NsxLibDhcpRelayProfile)

    def test_server_ips(self):
        fake_prf = test_constants.FAKE_RELAY_PROFILE.copy()
        relay_profile = self.get_mocked_resource()
        with mock.patch.object(relay_profile.client, 'url_get',
                               return_value=fake_prf):
            server_ips = relay_profile.get_server_ips(fake_prf['id'])
            self.assertEqual(1, len(server_ips))
            self.assertEqual(test_constants.FAKE_RELAY_SERVER,
                             server_ips[0])


class NsxLibBridgeClusterTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibBridgeClusterTestCase, self).setUp(
            core_resources.NsxLibBridgeCluster)


class NsxLibIpBlockSubnetTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibIpBlockSubnetTestCase, self).setUp(
            core_resources.NsxLibIpBlockSubnet)

    def test_list_all(self):
        if not self.resource:
            return
        mocked_resource = self.get_mocked_resource()
        block_id = '7'
        mocked_resource.list(block_id)
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s?block_id=%s' %
            (mocked_resource.uri_segment, block_id),
            headers=self.default_headers())


class NsxLibIpBlockTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibIpBlockTestCase, self).setUp(
            core_resources.NsxLibIpBlock)


class NsxLibFabricVirtualInterfaceTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibFabricVirtualInterfaceTestCase, self).setUp(
            core_resources.NsxLibFabricVirtualInterface)

    def test_get_by_owner_vm_id(self):
        mocked_resource = self.get_mocked_resource()
        vm_id = uuidutils.generate_uuid()
        mocked_resource.get_by_owner_vm_id(vm_id)
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s?owner_vm_id=%s' %
            (mocked_resource.uri_segment, vm_id),
            headers=self.default_headers())


class NsxLibFabricVirtualMachineTestCase(BaseTestResource):

    def setUp(self):
        super(NsxLibFabricVirtualMachineTestCase, self).setUp(
            core_resources.NsxLibFabricVirtualMachine)

    def test_get_by_display_name(self):
        mocked_resource = self.get_mocked_resource()
        display_name = 'some-vm-name'
        mocked_resource.get_by_display_name(display_name)
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s?display_name=%s' %
            (mocked_resource.uri_segment, display_name),
            headers=self.default_headers())


class LogicalDhcpServerTestCase(BaseTestResource):

    def setUp(self):
        super(LogicalDhcpServerTestCase, self).setUp(
            resources.LogicalDhcpServer)

    def test_update_empty_dhcp_server(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        ip = '1.1.1.1'

        with mock.patch.object(mocked_resource.client, "get", return_value={}):
            mocked_resource.update(server_uuid, server_ip=ip)
        body = {'ipv4_dhcp_server': {'dhcp_server_ip': ip}}

        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s' %
            (mocked_resource.uri_segment, server_uuid),
            data=jsonutils.dumps(body, sort_keys=True),
            headers=self.default_headers())

    def test_update_dhcp_server_new_val(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        ip = '1.1.1.1'
        domain_name = 'dummy'
        existing_server = {'ipv4_dhcp_server': {'domain_name': domain_name}}

        # add the server ip
        with mock.patch.object(mocked_resource.client, "get",
                               return_value=existing_server):
            mocked_resource.update(server_uuid, server_ip=ip)

        existing_server['ipv4_dhcp_server']['dhcp_server_ip'] = ip
        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s' %
            (mocked_resource.uri_segment, server_uuid),
            data=jsonutils.dumps(existing_server, sort_keys=True),
            headers=self.default_headers())

    def test_update_dhcp_server_replace_val(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        ip = '1.1.1.1'
        domain_name = 'dummy'
        existing_server = {'ipv4_dhcp_server': {'domain_name': domain_name,
                                                'dhcp_server_ip': ip}}

        # replace the server ip
        new_ip = '2.2.2.2'
        with mock.patch.object(mocked_resource.client, "get",
                               return_value=existing_server):
            mocked_resource.update(server_uuid, server_ip=new_ip)

        existing_server['ipv4_dhcp_server']['dhcp_server_ip'] = new_ip
        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s' %
            (mocked_resource.uri_segment, server_uuid),
            data=jsonutils.dumps(existing_server, sort_keys=True),
            headers=self.default_headers())

    def test_create_binding(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        mac = 'aa:bb:cc:dd:ee:ff'
        ip = '1.1.1.1'
        host = 'host'
        mocked_resource.create_binding(server_uuid, mac, ip, hostname=host)
        body = {
            'mac_address': mac,
            'ip_address': ip,
            'host_name': host,
        }
        test_client.assert_json_call(
            'post', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s/static-bindings' %
            (mocked_resource.uri_segment, server_uuid),
            data=jsonutils.dumps(body, sort_keys=True),
            headers=self.default_headers())

    def test_get_binding(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        binding_uuid = 'binding-uuid'
        mocked_resource.get_binding(server_uuid, binding_uuid)
        test_client.assert_json_call(
            'get', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s/static-bindings/%s' %
            (mocked_resource.uri_segment, server_uuid, binding_uuid),
            headers=self.default_headers())

    def test_update_binding(self):
        mocked_resource = self.get_mocked_resource()
        server_uuid = 'server-uuid'
        binding_uuid = 'binding-uuid'
        mac = 'aa:bb:cc:dd:ee:ff'
        new_mac = 'dd:bb:cc:dd:ee:ff'
        ip = '1.1.1.1'
        host = 'host'
        body = {
            'mac_address': mac,
            'ip_address': ip,
            'host_name': host,
        }
        with mock.patch.object(mocked_resource.client, "get",
                               return_value=body):
            mocked_resource.update_binding(server_uuid,
                                           binding_uuid,
                                           mac_address=new_mac)
        body['mac_address'] = new_mac
        test_client.assert_json_call(
            'put', mocked_resource,
            'https://1.2.3.4/api/v1/%s/%s/static-bindings/%s' %
            (mocked_resource.uri_segment, server_uuid, binding_uuid),
            data=jsonutils.dumps(body, sort_keys=True),
            headers=self.default_headers())


class NodeHttpServicePropertiesTestCase(BaseTestResource):

    def setUp(self):
        super(NodeHttpServicePropertiesTestCase, self).setUp(
            resources.NodeHttpServiceProperties)

    def test_get_resource(self):
        self.skipTest("The action is not supported by this resource")

    def test_list_all(self):
        self.skipTest("The action is not supported by this resource")

    def test_delete_resource(self):
        self.skipTest("The action is not supported by this resource")

    def test_get_rate_limit(self):
        mocked_resource = self.get_mocked_resource()
        rate_limit = 40
        body = {'service_properties': {'client_api_rate_limit': rate_limit}}
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.2.0'),\
            mock.patch.object(mocked_resource.client, "url_get",
                              return_value=body):
            result = mocked_resource.get_rate_limit()
            self.assertEqual(rate_limit, result)

    def test_update_rate_limit(self):
        mocked_resource = self.get_mocked_resource()
        old_rate_limit = 40
        new_rate_limit = 50
        body = {'service_properties': {
                'client_api_rate_limit': old_rate_limit}}
        with mock.patch("vmware_nsxlib.v3.NsxLib.get_version",
                        return_value='2.2.0'),\
            mock.patch.object(mocked_resource.client, "url_get",
                              return_value=body):
            mocked_resource.update_rate_limit(new_rate_limit)
            body['service_properties'][
                'client_api_rate_limit'] = new_rate_limit
            test_client.assert_json_call(
                'put', mocked_resource,
                'https://1.2.3.4/api/v1/node/services/http',
                data=jsonutils.dumps(body, sort_keys=True),
                headers=self.default_headers())
            test_client.assert_json_call(
                'post', mocked_resource,
                'https://1.2.3.4/api/v1/node/services/http?action=restart',
                headers=self.default_headers())


class DummyCachedResource(utils.NsxLibApiBase):

    @property
    def uri_segment(self):
        return 'XXX'

    @property
    def resource_type(self):
        return 'xxx'

    @property
    def use_cache_for_get(self):
        return True

    @property
    def cache_timeout(self):
        return 2


class ResourceCache(BaseTestResource):

    def setUp(self):
        super(ResourceCache, self).setUp(DummyCachedResource)

    def test_get_with_cache(self):
        mocked_resource = self.get_mocked_resource()
        fake_uuid = uuidutils.generate_uuid()
        # first call -> goes to the client
        mocked_resource.get(fake_uuid)
        self.assertEqual(1, test_client.mock_calls_count(
            'get', mocked_resource))

        # second call -> goes to cache
        mocked_resource.get(fake_uuid)
        self.assertEqual(1, test_client.mock_calls_count(
            'get', mocked_resource))

        # a different call -> goes to the client
        fake_uuid2 = uuidutils.generate_uuid()
        mocked_resource.get(fake_uuid2)
        self.assertEqual(2, test_client.mock_calls_count(
            'get', mocked_resource))

        # third call -> still goes to cache
        mocked_resource.get(fake_uuid)
        self.assertEqual(2, test_client.mock_calls_count(
            'get', mocked_resource))

        # after timeout -> goes to the client
        eventlet.sleep(2)
        mocked_resource.get(fake_uuid)
        self.assertEqual(3, test_client.mock_calls_count(
            'get', mocked_resource))

        # after delete -> goes to the client
        mocked_resource.delete(fake_uuid)
        mocked_resource.get(fake_uuid)
        self.assertEqual(4, test_client.mock_calls_count(
            'get', mocked_resource))

        # And from cache again
        mocked_resource.get(fake_uuid)
        self.assertEqual(4, test_client.mock_calls_count(
            'get', mocked_resource))

        # Update the entry. The get inside the update is from
        # the client too, because it must be current)
        mocked_resource._update_with_retry(fake_uuid, {})
        self.assertEqual(5, test_client.mock_calls_count(
            'get', mocked_resource))

        # after update -> goes to client
        mocked_resource.get(fake_uuid)
        self.assertEqual(6, test_client.mock_calls_count(
            'get', mocked_resource))



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import unittest

import mock

from vmware_nsxlib.tests.unit.v3 import nsxlib_testcase
from vmware_nsxlib import v3
from vmware_nsxlib.v3 import policy_constants
from vmware_nsxlib.v3 import policy_defs

TEST_TENANT = 'test'


class NsxPolicyLibTestCase(unittest.TestCase):

    def setUp(self, *args, **kwargs):
        super(NsxPolicyLibTestCase, self).setUp()

        nsxlib_config = nsxlib_testcase.get_default_nsxlib_config()
        self.policy_lib = v3.NsxPolicyLib(nsxlib_config)
        self.policy_api = self.policy_lib.policy_api

        self.maxDiff = None

    def _compare_def(self, expected_def, actual_def):
        # verify the resource definition class
        self.assertEqual(expected_def.__class__, actual_def.__class__)
        # verify the resource definition tenant
        self.assertEqual(expected_def.tenant, actual_def.tenant)
        # verify the resource definition values
        self.assertEqual(expected_def.get_obj_dict(),
                         actual_def.get_obj_dict())

    def assert_called_with_def(self, mock_api, expected_def, call_num=0):
        # verify the api was called
        mock_api.assert_called()
        actual_def = mock_api.call_args_list[call_num][0][0]
        self._compare_def(expected_def, actual_def)

    def assert_called_with_defs(self, mock_api, expected_defs, call_num=0):
        # verify the api & first resource definition
        self.assert_called_with_def(mock_api, expected_defs[0],
                                    call_num=call_num)
        # compare the 2nd resource definition class & values
        actual_def = mock_api.call_args_list[call_num][0][1]
        expected_def = expected_defs[1]
        self._compare_def(expected_def, actual_def)

    def assert_called_with_def_and_dict(self, mock_api,
                                        expected_def, expected_dict,
                                        call_num=0):
        # verify the api & resource definition
        self.assert_called_with_def(mock_api, expected_def,
                                    call_num=call_num)
        # compare the 2nd api parameter which is a dictionary
        actual_dict = mock_api.call_args_list[call_num][0][0].body
        self.assertEqual(expected_dict, actual_dict)


class TestPolicyDomain(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyDomain, self).setUp()
        self.resourceApi = self.policy_lib.domain

    def test_create_with_id(self):
        name = 'd1'
        description = 'desc'
        id = '111'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(name,
                                                 domain_id=id,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(domain_id=id,
                                                 name=name,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_create_without_id(self):
        name = 'd1'
        description = 'desc'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(name, description=description,
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(domain_id=mock.ANY,
                                                 name=name,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_delete(self):
        id = '111'
        with mock.patch.object(self.policy_api, "delete") as api_call:
            self.resourceApi.delete(id, tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(domain_id=id,
                                                 tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(domain_id=id,
                                                 tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 'd1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.DomainDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.DomainDef(domain_id=id,
                                                 tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'description': description}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)


class TestPolicyGroup(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyGroup, self).setUp()
        self.resourceApi = self.policy_lib.group

    def test_create_with_id(self):
        domain_id = '111'
        name = 'g1'
        description = 'desc'
        id = '222'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(name,
                                                 domain_id,
                                                 group_id=id,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                name=name,
                                                description=description,
                                                conditions=[],
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_create_without_id(self):
        domain_id = '111'
        name = 'g1'
        description = 'desc'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(name, domain_id,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=mock.ANY,
                                                name=name,
                                                description=description,
                                                conditions=[],
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_create_with_condition(self):
        domain_id = '111'
        name = 'g1'
        description = 'desc'
        cond_val = '123'
        cond_op = policy_constants.CONDITION_OP_EQUALS
        cond_member_type = policy_constants.CONDITION_MEMBER_VM
        cond_key = policy_constants.CONDITION_KEY_TAG
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(
                name, domain_id, description=description,
                cond_val=cond_val,
                cond_op=cond_op,
                cond_member_type=cond_member_type,
                cond_key=cond_key,
                tenant=TEST_TENANT)
            exp_cond = policy_defs.Condition(value=cond_val,
                                             key=cond_key,
                                             operator=cond_op,
                                             member_type=cond_member_type)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=mock.ANY,
                                                name=name,
                                                description=description,
                                                conditions=[exp_cond],
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_delete(self):
        domain_id = '111'
        id = '222'
        with mock.patch.object(self.policy_api, "delete") as api_call:
            self.resourceApi.delete(domain_id, id, tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        domain_id = '111'
        id = '222'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(domain_id, id, tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        domain_id = '111'
        name = 'g1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(domain_id, name,
                                               tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.GroupDef(domain_id, tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        domain_id = '111'
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(domain_id, tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        domain_id = '111'
        id = '222'
        name = 'new name'
        description = 'new desc'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as update_call:
            self.resourceApi.update(domain_id, id,
                                    name=name,
                                    description=description,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'description': description}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_update_condition(self):
        domain_id = '111'
        id = '222'
        cond_val = '123'
        with mock.patch.object(self.policy_api, "get",
                               return_value={}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update_condition(domain_id, id,
                                              cond_val=cond_val,
                                              tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                tenant=TEST_TENANT)
            exp_cond = {'resource_type': 'Condition',
                        'member_type': policy_constants.CONDITION_MEMBER_PORT,
                        'key': policy_constants.CONDITION_KEY_TAG,
                        'value': cond_val,
                        'operator': policy_constants.CONDITION_OP_EQUALS}
            expected_dict = {'expression': [exp_cond]}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_remove_condition(self):
        domain_id = '111'
        id = '222'
        old_cond = {'resource_type': 'Condition',
                    'member_type': policy_constants.CONDITION_MEMBER_PORT,
                    'key': policy_constants.CONDITION_KEY_TAG,
                    'value': 'abc',
                    'operator': policy_constants.CONDITION_OP_EQUALS}
        with mock.patch.object(self.policy_api, "get",
                               return_value={'expression': [old_cond]}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update_condition(domain_id, id,
                                              cond_val=None,
                                              tenant=TEST_TENANT)
            expected_def = policy_defs.GroupDef(domain_id=domain_id,
                                                group_id=id,
                                                tenant=TEST_TENANT)
            expected_dict = {'expression': []}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_get_realized(self):
        domain_id = 'd1'
        id = 'g1'
        ep_id = 'ef1'
        result = {'state': policy_constants.STATE_REALIZED}
        with mock.patch.object(
            self.policy_api, "get_by_path",
            return_value=result) as api_get:
            state = self.resourceApi.get_realized_state(
                domain_id, id, ep_id, tenant=TEST_TENANT)
            self.assertEqual(policy_constants.STATE_REALIZED, state)
            expected_path = policy_defs.REALIZED_STATE_GROUP % (
                TEST_TENANT, ep_id, domain_id, id)
            api_get.assert_called_once_with(expected_path)


class TestPolicyService(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyService, self).setUp()
        self.resourceApi = self.policy_lib.service

    def test_create(self):
        name = 's1'
        description = 'desc'
        protocol = policy_constants.TCP
        dest_ports = [81, 82]
        with mock.patch.object(self.policy_api,
                               "create_with_parent") as api_call:
            self.resourceApi.create_or_overwrite(name,
                                                 description=description,
                                                 protocol=protocol,
                                                 dest_ports=dest_ports,
                                                 tenant=TEST_TENANT)
            exp_srv_def = policy_defs.ServiceDef(service_id=mock.ANY,
                                                 name=name,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            exp_entry_def = policy_defs.L4ServiceEntryDef(
                service_id=mock.ANY,
                name=name,
                description=description,
                protocol=protocol,
                dest_ports=dest_ports,
                tenant=TEST_TENANT)
            self.assert_called_with_defs(
                api_call, [exp_srv_def, exp_entry_def])

    def test_delete(self):
        id = '111'
        with mock.patch.object(self.policy_api, "delete") as api_call,\
            mock.patch.object(self.policy_api, "get") as get_call:
            self.resourceApi.delete(id, tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 's1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.ServiceDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        with mock.patch.object(self.policy_api, "get",
                               return_value={}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'description': description,
                             'service_entries': []}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_update_entry(self):
        id = '111'
        protocol = 'udp'
        dest_ports = [555]
        service_entry_id = '222'
        service_entry = {'id': service_entry_id}

        with mock.patch.object(
            self.policy_api, "get",
            return_value={'service_entries': [service_entry]}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    protocol=protocol,
                                    dest_ports=dest_ports,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service entry only
            expected_entry_def = policy_defs.L4ServiceEntryDef(
                service_id=id,
                service_entry_id=service_entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': service_entry_id,
                                   'l4_protocol': protocol.upper(),
                                   'destination_ports': dest_ports}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict)

    def test_update_all(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        protocol = 'udp'
        dest_ports = [555]
        service_entry_id = '222'
        service_entry = {'id': service_entry_id}

        with mock.patch.object(
            self.policy_api, "get",
            return_value={'service_entries': [service_entry]}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call,\
            mock.patch.object(self.policy_api, "list",
                              return_value={'results': []}):
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    protocol=protocol,
                                    dest_ports=dest_ports,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service and entry (2 calls)
            expected_dict = {'display_name': name,
                             'description': description,
                             'service_entries': []}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

            expected_entry_def = policy_defs.L4ServiceEntryDef(
                service_id=id,
                service_entry_id=service_entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': service_entry_id,
                                   'display_name': name,
                                   'description': description,
                                   'l4_protocol': protocol.upper(),
                                   'destination_ports': dest_ports}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict,
                call_num=1)


class TestPolicyIcmpService(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyIcmpService, self).setUp()
        self.resourceApi = self.policy_lib.icmp_service

    def test_create(self):
        name = 's1'
        description = 'desc'
        icmp_type = 2
        with mock.patch.object(self.policy_api,
                               "create_with_parent") as api_call:
            self.resourceApi.create_or_overwrite(name,
                                                 description=description,
                                                 icmp_type=icmp_type,
                                                 tenant=TEST_TENANT)
            exp_srv_def = policy_defs.ServiceDef(service_id=mock.ANY,
                                                 name=name,
                                                 description=description,
                                                 tenant=TEST_TENANT)
            exp_entry_def = policy_defs.IcmpServiceEntryDef(
                service_id=mock.ANY,
                name=name,
                description=description,
                icmp_type=icmp_type,
                tenant=TEST_TENANT)
            self.assert_called_with_defs(
                api_call, [exp_srv_def, exp_entry_def])

    def test_delete(self):
        id = '111'
        with mock.patch.object(self.policy_api, "delete") as api_call,\
            mock.patch.object(self.policy_api, "get") as get_call:
            self.resourceApi.delete(id, tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 's1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.ServiceDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        with mock.patch.object(self.policy_api, "get",
                               return_value={}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'description': description,
                             'service_entries': []}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_update_entry(self):
        id = '111'
        icmp_code = 12
        service_entry_id = '222'
        service_entry = {'id': service_entry_id}

        with mock.patch.object(
            self.policy_api, "get",
            return_value={'service_entries': [service_entry]}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    icmp_code=icmp_code,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service entry only
            expected_entry_def = policy_defs.IcmpServiceEntryDef(
                service_id=id,
                service_entry_id=service_entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': service_entry_id,
                                   'icmp_code': icmp_code}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict)

    def test_update_all(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        version = 6
        icmp_type = 3
        icmp_code = 3
        service_entry_id = '222'
        service_entry = {'id': service_entry_id}

        with mock.patch.object(
            self.policy_api, "get",
            return_value={'service_entries': [service_entry]}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call,\
            mock.patch.object(self.policy_api, "list",
                              return_value={'results': []}):
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    version=version,
                                    icmp_type=icmp_type,
                                    icmp_code=icmp_code,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.ServiceDef(service_id=id,
                                                  tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service and entry (2 calls)
            expected_dict = {'display_name': name,
                             'description': description,
                             'service_entries': []}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

            expected_entry_def = policy_defs.IcmpServiceEntryDef(
                service_id=id,
                service_entry_id=service_entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': service_entry_id,
                                   'display_name': name,
                                   'description': description,
                                   'protocol': 'ICMPv6',
                                   'icmp_type': icmp_type,
                                   'icmp_code': icmp_code}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict,
                call_num=1)


class TestPolicyCommunicationProfile(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyCommunicationProfile, self).setUp()
        self.resourceApi = self.policy_lib.comm_profile

    def test_create(self):
        name = 'c1'
        description = 'desc'
        service_id = '333'
        action = 'DENY'
        with mock.patch.object(self.policy_api,
                               "create_with_parent") as api_call:
            self.resourceApi.create_or_overwrite(name, description=description,
                                                 services=[service_id],
                                                 action=action,
                                                 tenant=TEST_TENANT)
            exp_srv_def = policy_defs.CommunicationProfileDef(
                profile_id=mock.ANY,
                name=name,
                description=description,
                tenant=TEST_TENANT)
            exp_entry_def = policy_defs.CommunicationProfileEntryDef(
                profile_id=mock.ANY,
                name=name,
                description=description,
                services=[service_id],
                action=action,
                tenant=TEST_TENANT)
            self.assert_called_with_defs(
                api_call, [exp_srv_def, exp_entry_def])

    def test_delete(self):
        id = '111'
        with mock.patch.object(self.policy_api, "delete") as api_call,\
            mock.patch.object(self.policy_api, "get") as get_call:
            self.resourceApi.delete(id, tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationProfileDef(
                profile_id=id, tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationProfileDef(
                profile_id=id, tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 'c1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.CommunicationProfileDef(
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationProfileDef(
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        with mock.patch.object(self.policy_api, "get",
                               return_value={}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationProfileDef(
                profile_id=id, tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'description': description,
                             'communication_profile_entries': []}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_update_entry(self):
        id = '111'
        service_id = '333'
        action = 'deny'
        entry_id = '222'
        profile_entry = {'id': entry_id}
        entries_dict = {'communication_profile_entries': [profile_entry]}

        with mock.patch.object(
            self.policy_api, "get", return_value=entries_dict) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    services=[service_id],
                                    action=action,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.CommunicationProfileDef(
                profile_id=id, tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service entry only
            expected_entry_def = policy_defs.CommunicationProfileEntryDef(
                profile_id=id,
                profile_entry_id=entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': entry_id,
                                   'action': action.upper(),
                                   'services': [service_id]}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict)

    def test_update_all(self):
        id = '111'
        name = 'new name'
        description = 'new desc'
        service_id = '333'
        action = 'deny'
        entry_id = '222'
        profile_entry = {'id': entry_id}
        entries_dict = {'communication_profile_entries': [profile_entry]}

        with mock.patch.object(
            self.policy_api, "get", return_value=entries_dict) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    description=description,
                                    services=[service_id],
                                    action=action,
                                    tenant=TEST_TENANT)
            # get will be called for the entire service
            expected_def = policy_defs.CommunicationProfileDef(
                profile_id=id, tenant=TEST_TENANT)
            self.assert_called_with_def(get_call, expected_def)

            # update will be called for the service and entry (2 calls)
            expected_dict = {'display_name': name,
                             'description': description,
                             'communication_profile_entries': []}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

            expected_entry_def = policy_defs.CommunicationProfileEntryDef(
                profile_id=id,
                profile_entry_id=entry_id,
                tenant=TEST_TENANT)
            expected_entry_dict = {'id': entry_id,
                                   'display_name': name,
                                   'description': description,
                                   'action': action.upper(),
                                   'services': [service_id]}
            self.assert_called_with_def_and_dict(
                update_call, expected_entry_def, expected_entry_dict,
                call_num=1)


class TestPolicyCommunicationMap(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyCommunicationMap, self).setUp()
        self.resourceApi = self.policy_lib.comm_map

    def test_create(self):
        domain_id = '111'
        name = 'cm1'
        description = 'desc'
        source_group = 'g1'
        dest_group = 'g2'
        seq_num = 7
        profile_id = 'c1'
        list_return_value = {'results': [{'sequence_number': 1}]}
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call,\
            mock.patch.object(self.policy_api, "list",
                              return_value=list_return_value):
            self.resourceApi.create_or_overwrite(name, domain_id,
                                                 description=description,
                                                 sequence_number=seq_num,
                                                 profile_id=profile_id,
                                                 source_groups=[source_group],
                                                 dest_groups=[dest_group],
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=mock.ANY,
                name=name,
                description=description,
                sequence_number=seq_num,
                profile_id=profile_id,
                source_groups=[source_group],
                dest_groups=[dest_group],
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_create_first_seqnum(self):
        domain_id = '111'
        name = 'cm1'
        description = 'desc'
        source_group = 'g1'
        dest_group = 'g2'
        profile_id = 'c1'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call, \
            mock.patch.object(self.resourceApi, "list", return_value=[]):
            self.resourceApi.create_or_overwrite(name, domain_id,
                                                 description=description,
                                                 profile_id=profile_id,
                                                 source_groups=[source_group],
                                                 dest_groups=[dest_group],
                                                 tenant=TEST_TENANT)

            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=mock.ANY,
                name=name,
                description=description,
                sequence_number=1,
                profile_id=profile_id,
                source_groups=[source_group],
                dest_groups=[dest_group],
                tenant=TEST_TENANT)

            self.assert_called_with_def(api_call, expected_def)

    def test_create_without_seqnum(self):
        domain_id = '111'
        name = 'cm1'
        description = 'desc'
        source_group = 'g1'
        dest_group = 'g2'
        profile_id = 'c1'
        with mock.patch.object(self.policy_api,
                               "create_with_parent") as api_call, \
            mock.patch.object(self.resourceApi, "_get_last_seq_num",
                              return_value=-1):
            self.resourceApi.create_or_overwrite(name, domain_id,
                                                 description=description,
                                                 profile_id=profile_id,
                                                 source_groups=[source_group],
                                                 dest_groups=[dest_group],
                                                 tenant=TEST_TENANT)

            expected_map_def = policy_defs.CommunicationMapDef(
                domain_id=domain_id,
                tenant=TEST_TENANT)

            expected_entry_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=mock.ANY,
                name=name,
                description=description,
                sequence_number=1,
                profile_id=profile_id,
                source_groups=[source_group],
                dest_groups=[dest_group],
                tenant=TEST_TENANT)

            self.assert_called_with_defs(
                api_call,
                [expected_map_def, expected_entry_def])

    def test_delete(self):
        domain_id = '111'
        id = '222'
        with mock.patch.object(self.policy_api, "delete") as api_call:
            self.resourceApi.delete(domain_id, id, tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=id,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        domain_id = '111'
        id = '222'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(domain_id, id, tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=id,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        domain_id = '111'
        name = 'cm1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(domain_id, name,
                                               tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        domain_id = '111'
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(domain_id, tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        domain_id = '111'
        id = '222'
        name = 'new name'
        description = 'new desc'
        source_group = 'ng1'
        dest_group = 'ng2'
        profile_id = 'nc1'
        with mock.patch.object(self.policy_api, "get",
                               return_value={}) as get_call,\
            mock.patch.object(self.policy_api,
                              "create_or_update") as update_call:
            self.resourceApi.update(domain_id, id,
                                    name=name,
                                    description=description,
                                    profile_id=profile_id,
                                    source_groups=[source_group],
                                    dest_groups=[dest_group],
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.CommunicationMapEntryDef(
                domain_id=domain_id,
                map_id=id,
                tenant=TEST_TENANT)
            sgroup_path = "/%s/domains/%s/groups/%s" % (
                TEST_TENANT, domain_id, source_group)
            dgroup_path = "/%s/domains/%s/groups/%s" % (
                TEST_TENANT, domain_id, dest_group)
            profile_path = "/%s/communication-profiles/%s" % (
                TEST_TENANT, profile_id)
            expected_dict = {'display_name': name,
                             'description': description,
                             'communication_profile_path': profile_path,
                             'source_groups': [sgroup_path],
                             'destination_groups': [dgroup_path]}
            self.assert_called_with_def(get_call, expected_def)
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_get_realized(self):
        domain_id = 'd1'
        ep_id = 'ef1'
        result = {'state': policy_constants.STATE_REALIZED}
        with mock.patch.object(
            self.policy_api, "get_by_path",
            return_value=result) as api_get:
            state = self.resourceApi.get_realized_state(
                domain_id, ep_id, tenant=TEST_TENANT)
            self.assertEqual(policy_constants.STATE_REALIZED, state)
            expected_path = policy_defs.REALIZED_STATE_COMM_MAP % (
                TEST_TENANT, ep_id, domain_id)
            api_get.assert_called_once_with(expected_path)


class TestPolicyEnforcementPoint(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyEnforcementPoint, self).setUp()
        self.resourceApi = self.policy_lib.enforcement_point

    def test_create(self):
        name = 'ep'
        description = 'desc'
        ip_address = '1.1.1.1'
        username = 'admin'
        password = 'zzz'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(
                name, description=description,
                ip_address=ip_address,
                username=username,
                password=password,
                tenant=TEST_TENANT)

            expected_def = policy_defs.EnforcementPointDef(
                ep_id=mock.ANY,
                name=name,
                description=description,
                ip_address=ip_address,
                username=username,
                password=password,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_delete(self):
        id = '111'
        with mock.patch.object(self.policy_api, "delete") as api_call:
            self.resourceApi.delete(id, tenant=TEST_TENANT)
            expected_def = policy_defs.EnforcementPointDef(ep_id=id,
                                                           tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, tenant=TEST_TENANT)
            expected_def = policy_defs.EnforcementPointDef(ep_id=id,
                                                           tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 'ep1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.EnforcementPointDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(tenant=TEST_TENANT)
            expected_def = policy_defs.EnforcementPointDef(tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        username = 'admin'
        password = 'zzz'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    username=username,
                                    password=password,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.EnforcementPointDef(ep_id=id,
                                                           tenant=TEST_TENANT)
            expected_dict = {'display_name': name,
                             'username': username,
                             'password': password}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)

    def test_get_realized(self):
        ep_id = 'ef1'
        result = {'state': policy_constants.STATE_REALIZED}
        with mock.patch.object(
            self.policy_api, "get_by_path",
            return_value=result) as api_get:
            state = self.resourceApi.get_realized_state(
                ep_id, tenant=TEST_TENANT)
            self.assertEqual(policy_constants.STATE_REALIZED, state)
            expected_path = policy_defs.REALIZED_STATE_EF % (
                TEST_TENANT, ep_id)
            api_get.assert_called_once_with(expected_path)


class TestPolicyDeploymentMap(NsxPolicyLibTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPolicyDeploymentMap, self).setUp()
        self.resourceApi = self.policy_lib.deployment_map

    def test_create(self):
        name = 'map1'
        description = 'desc'
        domain_id = 'domain1'
        ep_id = 'ep1'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as api_call:
            self.resourceApi.create_or_overwrite(name,
                                                 description=description,
                                                 ep_id=ep_id,
                                                 domain_id=domain_id,
                                                 tenant=TEST_TENANT)
            expected_def = policy_defs.DeploymentMapDef(
                map_id=mock.ANY,
                name=name,
                description=description,
                ep_id=ep_id,
                domain_id=domain_id,
                tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_delete(self):
        id = '111'
        domain_id = 'domain1'
        with mock.patch.object(self.policy_api, "delete") as api_call:
            self.resourceApi.delete(id, domain_id=domain_id,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.DeploymentMapDef(map_id=id,
                                                        domain_id=domain_id,
                                                        tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get(self):
        id = '111'
        domain_id = 'domain1'
        with mock.patch.object(self.policy_api, "get") as api_call:
            self.resourceApi.get(id, domain_id=domain_id, tenant=TEST_TENANT)
            expected_def = policy_defs.DeploymentMapDef(map_id=id,
                                                        domain_id=domain_id,
                                                        tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_get_by_name(self):
        name = 'ep1'
        domain_id = 'domain1'
        with mock.patch.object(
            self.policy_api, "list",
            return_value={'results': [{'display_name': name}]}) as api_call:
            obj = self.resourceApi.get_by_name(name, domain_id=domain_id,
                                               tenant=TEST_TENANT)
            self.assertIsNotNone(obj)
            expected_def = policy_defs.DeploymentMapDef(domain_id=domain_id,
                                                        tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_list(self):
        domain_id = 'domain1'
        with mock.patch.object(self.policy_api, "list") as api_call:
            self.resourceApi.list(domain_id=domain_id, tenant=TEST_TENANT)
            expected_def = policy_defs.DeploymentMapDef(domain_id=domain_id,
                                                        tenant=TEST_TENANT)
            self.assert_called_with_def(api_call, expected_def)

    def test_update(self):
        id = '111'
        name = 'new name'
        domain_id = 'domain2'
        ep_id = 'ep2'
        with mock.patch.object(self.policy_api,
                               "create_or_update") as update_call:
            self.resourceApi.update(id,
                                    name=name,
                                    ep_id=ep_id,
                                    domain_id=domain_id,
                                    tenant=TEST_TENANT)
            expected_def = policy_defs.DeploymentMapDef(map_id=id,
                                                        tenant=TEST_TENANT)
            domain_path = "/%s/domains/%s" % (TEST_TENANT, domain_id)
            ep_path = ("/%s/deployment-zones/default/"
                       "enforcement-points/%s" % (TEST_TENANT, ep_id))
            expected_dict = {'display_name': name,
                             'enforcement_point_paths': [ep_path],
                             'domain_path': domain_path}
            self.assert_called_with_def_and_dict(
                update_call, expected_def, expected_dict)



# Copyright 2015 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import abc
import contextlib
import copy
import datetime
import inspect
import itertools
import logging
import re

import eventlet
from eventlet import greenpool
from eventlet import pools
import OpenSSL
from oslo_log import log
from oslo_service import loopingcall
import requests
from requests import adapters
from requests import exceptions as requests_exceptions
import six
import six.moves.urllib.parse as urlparse

from vmware_nsxlib._i18n import _
from vmware_nsxlib.v3 import client as nsx_client
from vmware_nsxlib.v3 import exceptions


LOG = log.getLogger(__name__)

# disable warning message for each HTTP retry
logging.getLogger(
    "requests.packages.urllib3.connectionpool").setLevel(logging.ERROR)


@six.add_metaclass(abc.ABCMeta)
class AbstractHTTPProvider(object):
    """Interface for providers of HTTP connections.

    which are responsible for creating and validating connections
    for their underlying HTTP support.
    """
    @property
    def default_scheme(self):
        return 'https'

    @abc.abstractproperty
    def provider_id(self):
        """A unique string name for this provider."""
        pass

    @abc.abstractmethod
    def validate_connection(self, cluster_api, endpoint, conn):
        """Validate the said connection for the given endpoint and cluster."""
        pass

    @abc.abstractmethod
    def new_connection(self, cluster_api, provider):
        """Create a new http connection.

        Create a new http connection for the said cluster and
        cluster provider. The actual connection should duck type
        requests.Session http methods (get(), put(), etc.).
        """
        pass

    @abc.abstractmethod
    def is_connection_exception(self, exception):
        """Determine if the given exception is related to connection failure.

        Return True if it's a connection exception and False otherwise.
        """


class TimeoutSession(requests.Session):
    """Extends requests.Session to support timeout at the session level."""

    def __init__(self, timeout, read_timeout):
        self.timeout = timeout
        self.read_timeout = read_timeout
        self.cert_provider = None
        super(TimeoutSession, self).__init__()

    @property
    def cert_provider(self):
        return self._cert_provider

    @cert_provider.setter
    def cert_provider(self, value):
        self._cert_provider = value

    # wrapper timeouts at the session level
    # see: https://goo.gl/xNk7aM
    def request(self, *args, **kwargs):
        def request_with_retry_on_ssl_error(*args, **kwargs):
            try:
                return super(TimeoutSession, self).request(*args, **kwargs)
            except OpenSSL.SSL.Error:
                # This can happen when connection tries to access certificate
                # file it was opened with (renegotiation?)
                # Proper way to solve this would be to pass in-memory cert
                # to ssl C code.
                # Retrying here works around the problem
                return super(TimeoutSession, self).request(*args, **kwargs)

        def get_cert_provider():
            if inspect.isclass(self._cert_provider):
                # If client provided certificate provider as a class,
                # we spawn an instance here
                return self._cert_provider()
            return self._cert_provider

        if 'timeout' not in kwargs:
            kwargs['timeout'] = (self.timeout, self.read_timeout)
        if not self.cert_provider:
            # No client certificate needed
            return super(TimeoutSession, self).request(*args, **kwargs)

        if self.cert is not None:
            # Recursive call - shouldn't happen
            return request_with_retry_on_ssl_error(*args, **kwargs)

        # The following with statement allows for preparing certificate and
        # private key file and dispose it at the end of request
        # (since PK is sensitive information, immediate disposal is
        # important).
        # It would be optimal to populate certificate once per connection,
        # per request. Unfortunately requests library verifies cert file
        # existence regardless of whether certificate is going to be used
        # for this request.
        # Optimal solution for this would be to expose certificate as variable
        # and not as a file to the SSL library
        with get_cert_provider() as provider:
            self.cert = provider.filename()
            try:
                ret = request_with_retry_on_ssl_error(*args, **kwargs)
            except Exception as e:
                self.cert = None
                raise e

            self.cert = None

        return ret


class NSXRequestsHTTPProvider(AbstractHTTPProvider):
    """Concrete implementation of AbstractHTTPProvider.

    using requests.Session() as the underlying connection.
    """

    SESSION_CREATE_URL = '/api/session/create'
    COOKIE_FIELD = 'Cookie'
    SET_COOKIE_FIELD = 'Set-Cookie'
    XSRF_TOKEN = 'X-XSRF-TOKEN'
    JSESSIONID = 'JSESSIONID'

    @property
    def provider_id(self):
        return "%s-%s" % (requests.__title__, requests.__version__)

    def validate_connection(self, cluster_api, endpoint, conn):
        client = nsx_client.NSX3Client(
            conn, url_prefix=endpoint.provider.url,
            url_path_base=cluster_api.nsxlib_config.url_base,
            default_headers=conn.default_headers)
        keepalive_section = cluster_api.nsxlib_config.keepalive_section
        result = client.get(keepalive_section, silent=True)
        # If keeplive section returns a list, it is assumed to be non-empty
        if not result or result.get('result_count', 1) <= 0:
            msg = _("No %(section)s found "
                    "for '%(url)s'") % {'section': keepalive_section,
                                        'url': endpoint.provider.url}
            LOG.warning(msg)
            raise exceptions.ResourceNotFound(
                manager=endpoint.provider.url, operation=msg)

    def new_connection(self, cluster_api, provider):
        config = cluster_api.nsxlib_config
        session = TimeoutSession(config.http_timeout,
                                 config.http_read_timeout)
        if config.client_cert_provider:
            session.cert_provider = config.client_cert_provider
        else:
            session.auth = (provider.username, provider.password)

        # NSX v3 doesn't use redirects
        session.max_redirects = 0

        session.verify = not config.insecure
        if session.verify and provider.ca_file:
            # verify using the said ca bundle path
            session.verify = provider.ca_file

        # we are pooling with eventlet in the cluster class
        adapter = adapters.HTTPAdapter(
            pool_connections=1, pool_maxsize=1,
            max_retries=config.retries,
            pool_block=False)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        self.get_default_headers(session, provider,
                                 config.allow_overwrite_header)

        return session

    def is_connection_exception(self, exception):
        return isinstance(exception, requests_exceptions.ConnectionError)

    def get_default_headers(self, session, provider, allow_overwrite_header):
        """Get the default headers that should be added to future requests"""
        session.default_headers = {}

        # Perform the initial session create and get the relevant jsessionid &
        # X-XSRF-TOKEN for future requests
        req_data = ''
        if not session.cert_provider:
            # With client certificate authentication, username and password
            # may not be provided.
            # If provided, backend treats these credentials as authentication
            # and ignores client cert as principal identity indication.
            req_data = 'j_username=%s&j_password=%s' % (provider.username,
                                                        provider.password)
        req_headers = {'Accept': 'application/json',
                       'Content-Type': 'application/x-www-form-urlencoded'}
        # Cannot use the certificate at this stage, because it is used for
        # the certificate generation
        resp = session.request('post', provider.url + self.SESSION_CREATE_URL,
                               data=req_data, headers=req_headers)
        if resp.status_code != 200:
            LOG.error("Session create failed for endpoint %s", provider.url)
            # this will later cause the endpoint to be Down
        else:
            for header_name in resp.headers:
                if self.SET_COOKIE_FIELD.lower() == header_name.lower():
                    m = re.match('%s=.*?\;' % self.JSESSIONID,
                                 resp.headers[header_name])
                    if m:
                        session.default_headers[self.COOKIE_FIELD] = m.group()
                if self.XSRF_TOKEN.lower() == header_name.lower():
                    session.default_headers[self.XSRF_TOKEN] = resp.headers[
                        header_name]
            LOG.info("Session create succeeded for endpoint %(url)s with "
                     "headers %(hdr)s",
                     {'url': provider.url, 'hdr': session.default_headers})

        # Add allow-overwrite if configured
        if allow_overwrite_header:
            session.default_headers['X-Allow-Overwrite'] = 'true'


class ClusterHealth(object):
    """Indicator of overall cluster health.

    with respect to the connectivity of the clusters managed endpoints.
    """
    # all endpoints are UP
    GREEN = 'GREEN'
    # at least 1 endpoint is UP, but 1 or more are DOWN
    ORANGE = 'ORANGE'
    # all endpoints are DOWN
    RED = 'RED'


class EndpointState(object):
    """Tracks the connectivity state for a said endpoint."""
    # no UP or DOWN state recorded yet
    INITIALIZED = 'INITIALIZED'
    # endpoint has been validate and is good
    UP = 'UP'
    # endpoint can't be reached or validated
    DOWN = 'DOWN'


class Provider(object):
    """Data holder for a provider

    Which has a unique id a connection URL, and the credential details.
    """

    def __init__(self, provider_id, provider_url, username, password, ca_file):
        self.id = provider_id
        self.url = provider_url
        self.username = username
        self.password = password
        self.ca_file = ca_file

    def __str__(self):
        return str(self.url)


class Endpoint(object):
    """A single NSX manager endpoint (host).

    A single NSX manager endpoint (host) which includes
    related information such as the endpoint's provider,
    state, etc.. A pool is used to hold connections to the
    endpoint which are doled out when proxying HTTP methods
    to the underlying connections.
    """

    def __init__(self, provider, pool):
        self.provider = provider
        self.pool = pool
        self._state = EndpointState.INITIALIZED
        self._last_updated = datetime.datetime.now()

    def regenerate_pool(self):
        self.pool = pools.Pool(min_size=self.pool.min_size,
                               max_size=self.pool.max_size,
                               order_as_stack=True,
                               create=self.pool.create)

    @property
    def last_updated(self):
        return self._last_updated

    @property
    def state(self):
        return self._state

    def set_state(self, state):
        if self.state != state:
            LOG.info("Endpoint '%(ep)s' changing from state"
                     " '%(old)s' to '%(new)s'",
                     {'ep': self.provider,
                      'old': self.state,
                      'new': state})
        old_state = self._state
        self._state = state

        self._last_updated = datetime.datetime.now()

        return old_state

    def __str__(self):
        return "[%s] %s" % (self.state, self.provider)


class EndpointConnection(object):
    """Simple data holder

    Which contains an endpoint and a connection for that endpoint.
    """

    def __init__(self, endpoint, connection):
        self.endpoint = endpoint
        self.connection = connection


class ClusteredAPI(object):
    """Duck types the major HTTP based methods of a requests.Session

    Such as get(), put(), post(), etc.
    and transparently proxies those calls to one of
    its managed NSX manager endpoints.
    """
    _HTTP_VERBS = ['get', 'delete', 'head', 'put', 'post', 'patch', 'create']

    def __init__(self, providers,
                 http_provider,
                 min_conns_per_pool=1,
                 max_conns_per_pool=500,
                 keepalive_interval=33):

        self._http_provider = http_provider
        self._keepalive_interval = keepalive_interval

        def _init_cluster(*args, **kwargs):
            self._init_endpoints(providers,
                                 min_conns_per_pool, max_conns_per_pool)

        _init_cluster()

        # keep this internal method for reinitialize upon fork
        # for api workers to ensure each process has its own keepalive
        # loops + state
        self._reinit_cluster = _init_cluster

    def _init_endpoints(self, providers,
                        min_conns_per_pool, max_conns_per_pool):
        LOG.debug("Initializing API endpoints")

        def _create_conn(p):
            def _conn():
                # called when a pool needs to create a new connection
                return self._http_provider.new_connection(self, p)

            return _conn

        self._endpoints = {}
        for provider in providers:
            pool = pools.Pool(
                min_size=min_conns_per_pool,
                max_size=max_conns_per_pool,
                order_as_stack=True,
                create=_create_conn(provider))

            endpoint = Endpoint(provider, pool)
            self._endpoints[provider.id] = endpoint

        # service requests using round robin
        self._endpoint_schedule = itertools.cycle(self._endpoints.values())

        # duck type to proxy http invocations
        for method in ClusteredAPI._HTTP_VERBS:
            setattr(self, method, self._proxy_stub(method))

        conns = greenpool.GreenPool()
        for endpoint in self._endpoints.values():
            conns.spawn(self._validate, endpoint)
        eventlet.sleep(0)
        while conns.running():
            if (self.health == ClusterHealth.GREEN
                    or self.health == ClusterHealth.ORANGE):
                # only wait for 1 or more endpoints to reduce init time
                break
            eventlet.sleep(0.5)

        for endpoint in self._endpoints.values():
            # dynamic loop for each endpoint to ensure connectivity
            loop = loopingcall.DynamicLoopingCall(
                self._endpoint_keepalive, endpoint)
            loop.start(initial_delay=self._keepalive_interval,
                       periodic_interval_max=self._keepalive_interval,
                       stop_on_exception=False)

        LOG.debug("Done initializing API endpoint(s). "
                  "API cluster health: %s", self.health)

    def _endpoint_keepalive(self, endpoint):
        delta = datetime.datetime.now() - endpoint.last_updated
        if delta.seconds >= self._keepalive_interval:
            # TODO(boden): backoff on validation failure
            self._validate(endpoint)
            return self._keepalive_interval
        return self._keepalive_interval - delta.seconds

    @property
    def providers(self):
        return [ep.provider for ep in self._endpoints.values()]

    @property
    def endpoints(self):
        return copy.copy(self._endpoints)

    @property
    def http_provider(self):
        return self._http_provider

    @property
    def health(self):
        down = 0
        up = 0
        for endpoint in self._endpoints.values():
            if endpoint.state != EndpointState.UP:
                down += 1
            else:
                up += 1

        if down == len(self._endpoints):
            return ClusterHealth.RED
        return (ClusterHealth.GREEN
                if up == len(self._endpoints)
                else ClusterHealth.ORANGE)

    def _validate(self, endpoint):
        try:
            with endpoint.pool.item() as conn:
                self._http_provider.validate_connection(self, endpoint, conn)
                endpoint.set_state(EndpointState.UP)
        except exceptions.ClientCertificateNotTrusted:
            LOG.warning("Failed to validate API cluster endpoint "
                        "'%(ep)s' due to untrusted client certificate",
                        {'ep': endpoint})
            # regenerate connection pool based on new certificate
            endpoint.regenerate_pool()
        except exceptions.BadXSRFToken:
            LOG.warning("Failed to validate API cluster endpoint "
                        "'%(ep)s' due to expired XSRF token",
                        {'ep': endpoint})
            # regenerate connection pool based on token
            endpoint.regenerate_pool()
        except Exception as e:
            endpoint.set_state(EndpointState.DOWN)
            LOG.warning("Failed to validate API cluster endpoint "
                        "'%(ep)s' due to: %(err)s",
                        {'ep': endpoint, 'err': e})

    def _select_endpoint(self):
        # check for UP state until exhausting all endpoints
        seen, total = 0, len(self._endpoints.values())
        while seen < total:
            endpoint = next(self._endpoint_schedule)
            if endpoint.state == EndpointState.UP:
                return endpoint
            seen += 1

    def endpoint_for_connection(self, conn):
        # check all endpoint pools
        for endpoint in self._endpoints.values():
            if (conn in endpoint.pool.channel.queue or
                    conn in endpoint.pool.free_items):
                return endpoint

    @property
    def cluster_id(self):
        return ','.join([str(ep.provider.url)
                         for ep in self._endpoints.values()])

    @contextlib.contextmanager
    def connection(self):
        with self.endpoint_connection() as conn_data:
            yield conn_data.connection

    @contextlib.contextmanager
    def endpoint_connection(self):
        endpoint = self._select_endpoint()
        if not endpoint:
            LOG.debug("All endpoints down for: %s" %
                      [str(ep) for ep in self._endpoints.values()])
            # all endpoints are DOWN and will have their next
            # state updated as per _endpoint_keepalive()
            raise exceptions.ServiceClusterUnavailable(
                cluster_id=self.cluster_id)

        if endpoint.pool.free() == 0:
            LOG.info("API endpoint %(ep)s at connection "
                     "capacity %(max)s and has %(waiting)s waiting",
                     {'ep': endpoint,
                      'max': endpoint.pool.max_size,
                      'waiting': endpoint.pool.waiting()})
        # pool.item() will wait if pool has 0 free
        with endpoint.pool.item() as conn:
            yield EndpointConnection(endpoint, conn)

    def _proxy_stub(self, proxy_for):
        def _call_proxy(url, *args, **kwargs):
            return self._proxy(proxy_for, url, *args, **kwargs)
        return _call_proxy

    def _proxy(self, proxy_for, uri, *args, **kwargs):
        # proxy http request call to an avail endpoint
        with self.endpoint_connection() as conn_data:
            conn = conn_data.connection
            endpoint = conn_data.endpoint

            # http conn must support requests style interface
            do_request = getattr(conn, proxy_for)

            if not uri.startswith('/'):
                uri = "/%s" % uri
            url = "%s%s" % (endpoint.provider.url, uri)
            try:
                LOG.debug("API cluster proxy %s %s to %s",
                          proxy_for.upper(), uri, url)
                # Add the connection default headers
                if conn.default_headers:
                    kwargs['headers'] = kwargs.get('headers', {})
                    kwargs['headers'].update(conn.default_headers)

                # call the actual connection method to do the
                # http request/response over the wire
                response = do_request(url, *args, **kwargs)
                endpoint.set_state(EndpointState.UP)

                return response
            except Exception as e:
                LOG.warning("Request failed due to: %s", e)
                if not self._http_provider.is_connection_exception(e):
                    # only trap and retry connection errors
                    raise e
                endpoint.set_state(EndpointState.DOWN)
                LOG.debug("Connection to %s failed, checking additional "
                          "endpoints" % url)
                # retry until exhausting endpoints
                return self._proxy(proxy_for, uri, *args, **kwargs)


class NSXClusteredAPI(ClusteredAPI):
    """Extends ClusteredAPI to get conf values and setup the NSXv3 cluster."""

    def __init__(self, nsxlib_config):
        self.nsxlib_config = nsxlib_config

        self._http_provider = (nsxlib_config.http_provider or
                               NSXRequestsHTTPProvider())

        super(NSXClusteredAPI, self).__init__(
            self._build_conf_providers(),
            self._http_provider,
            max_conns_per_pool=self.nsxlib_config.concurrent_connections,
            keepalive_interval=self.nsxlib_config.conn_idle_timeout)

        LOG.debug("Created NSX clustered API with '%s' "
                  "provider", self._http_provider.provider_id)

    def _build_conf_providers(self):

        def _schemed_url(uri):
            uri = uri.strip('/')
            return urlparse.urlparse(
                uri if uri.startswith('http') else
                "%s://%s" % (self._http_provider.default_scheme, uri))

        conf_urls = self.nsxlib_config.nsx_api_managers[:]
        urls = []
        providers = []
        provider_index = -1
        for conf_url in conf_urls:
            provider_index += 1
            conf_url = _schemed_url(conf_url)
            if conf_url in urls:
                LOG.warning("'%s' already defined in configuration file. "
                            "Skipping.", urlparse.urlunparse(conf_url))
                continue
            urls.append(conf_url)
            providers.append(
                Provider(
                    conf_url.netloc,
                    urlparse.urlunparse(conf_url),
                    self.nsxlib_config.username(provider_index),
                    self.nsxlib_config.password(provider_index),
                    self.nsxlib_config.ca_file(provider_index)))
        return providers
